
# -*- coding: utf-8 -*-
from gitlint.tests.base import BaseTestCase
from gitlint.rules import RuleViolation
from gitlint.contrib.rules.conventional_commit import ConventionalCommit
from gitlint.config import LintConfig


class ContribConventionalCommitTests(BaseTestCase):

    def test_enable(self):
        # Test that rule can be enabled in config
        for rule_ref in ['CT1', 'contrib-title-conventional-commits']:
            config = LintConfig()
            config.contrib = [rule_ref]
            self.assertIn(ConventionalCommit(), config.rules)

    def test_conventional_commits(self):
        rule = ConventionalCommit()

        # No violations when using a correct type and format
        for type in ["fix", "feat", "chore", "docs", "style", "refactor", "perf", "test", "revert", "ci", "build"]:
            violations = rule.validate(type + ": föo", None)
            self.assertListEqual([], violations)

        # assert violation on wrong type
        expected_violation = RuleViolation("CT1", "Title does not start with one of fix, feat, chore, docs,"
                                                  " style, refactor, perf, test, revert, ci, build", "bår: foo")
        violations = rule.validate("bår: foo", None)
        self.assertListEqual([expected_violation], violations)

        # assert violation when use strange chars after correct type
        expected_violation = RuleViolation("CT1", "Title does not start with one of fix, feat, chore, docs,"
                                                  " style, refactor, perf, test, revert, ci, build",
                                                  "feat_wrong_chars: föo")
        violations = rule.validate("feat_wrong_chars: föo", None)
        self.assertListEqual([expected_violation], violations)

        # assert violation when use strange chars after correct type
        expected_violation = RuleViolation("CT1", "Title does not start with one of fix, feat, chore, docs,"
                                                  " style, refactor, perf, test, revert, ci, build",
                                                  "feat_wrong_chars(scope): föo")
        violations = rule.validate("feat_wrong_chars(scope): föo", None)
        self.assertListEqual([expected_violation], violations)

        # assert violation on wrong format
        expected_violation = RuleViolation("CT1", "Title does not follow ConventionalCommits.org format "
                                                  "'type(optional-scope): description'", "fix föo")
        violations = rule.validate("fix föo", None)
        self.assertListEqual([expected_violation], violations)

        # assert no violation when use ! for breaking changes without scope
        violations = rule.validate("feat!: föo", None)
        self.assertListEqual([], violations)

        # assert no violation when use ! for breaking changes with scope
        violations = rule.validate("fix(scope)!: föo", None)
        self.assertListEqual([], violations)

        # assert no violation when adding new type
        rule = ConventionalCommit({'types': ["föo", "bär"]})
        for typ in ["föo", "bär"]:
            violations = rule.validate(typ + ": hür dur", None)
            self.assertListEqual([], violations)

        # assert violation when using incorrect type when types have been reconfigured
        violations = rule.validate("fix: hür dur", None)
        expected_violation = RuleViolation("CT1", "Title does not start with one of föo, bär", "fix: hür dur")
        self.assertListEqual([expected_violation], violations)

        # assert no violation when adding new type named with numbers
        rule = ConventionalCommit({'types': ["föo123", "123bär"]})
        for typ in ["föo123", "123bär"]:
            violations = rule.validate(typ + ": hür dur", None)
            self.assertListEqual([], violations)



# -*- coding: utf-8 -*-

from gitlint.rules import CommitRule, RuleViolation
from gitlint.options import IntOption, ListOption

"""
Full details on user-defined rules: https://jorisroovers.com/gitlint/user_defined_rules

The classes below are examples of user-defined CommitRules. Commit rules are gitlint rules that
act on the entire commit at once. Once the rules are discovered, gitlint will automatically take care of applying them
to the entire commit. This happens exactly once per commit.

A CommitRule contrasts with a LineRule (see examples/my_line_rules.py) in that a commit rule is only applied once on
an entire commit. This allows commit rules to implement more complex checks that span multiple lines and/or checks
that should only be done once per gitlint run.

While every LineRule can be implemented as a CommitRule, it's usually easier and more concise to go with a LineRule if
that fits your needs.
"""


class BodyMaxLineCount(CommitRule):
    # A rule MUST have a human friendly name
    name = "body-max-line-count"

    # A rule MUST have a *unique* id, we recommend starting with UC (for User-defined Commit-rule).
    id = "UC1"

    # A rule MAY have an option_spec if its behavior should be configurable.
    options_spec = [IntOption('max-line-count', 3, "Maximum body line count")]

    def validate(self, commit):
        self.log.debug("BodyMaxLineCount: This will be visible when running `gitlint --debug`")

        line_count = len(commit.message.body)
        max_line_count = self.options['max-line-count'].value
        if line_count > max_line_count:
            message = f"Body contains too many lines ({line_count} > {max_line_count})"
            return [RuleViolation(self.id, message, line_nr=1)]


class SignedOffBy(CommitRule):
    """ This rule will enforce that each commit contains a "Signed-off-by" line.
    We keep things simple here and just check whether the commit body contains a line that starts with "Signed-off-by".
    """

    # A rule MUST have a human friendly name
    name = "body-requires-signed-off-by"

    # A rule MUST have a *unique* id, we recommend starting with UC (for User-defined Commit-rule).
    id = "UC2"

    def validate(self, commit):
        self.log.debug("SignedOffBy: This will be visible when running `gitlint --debug`")

        for line in commit.message.body:
            if line.startswith("Signed-off-by"):
                return

        return [RuleViolation(self.id, "Body does not contain a 'Signed-off-by' line", line_nr=1)]


class BranchNamingConventions(CommitRule):
    """ This rule will enforce that a commit is part of a branch that meets certain naming conventions.
        See GitFlow for real-world example of this: https://nvie.com/posts/a-successful-git-branching-model/
    """

    # A rule MUST have a human friendly name
    name = "branch-naming-conventions"

    # A rule MUST have a *unique* id, we recommend starting with UC (for User-defined Commit-rule).
    id = "UC3"

    # A rule MAY have an option_spec if its behavior should be configurable.
    options_spec = [ListOption('branch-prefixes', ["feature/", "hotfix/", "release/"], "Allowed branch prefixes")]

    def validate(self, commit):
        self.log.debug("BranchNamingConventions: This line will be visible when running `gitlint --debug`")

        violations = []
        allowed_branch_prefixes = self.options['branch-prefixes'].value
        for branch in commit.branches:
            valid_branch_name = False

            for allowed_prefix in allowed_branch_prefixes:
                if branch.startswith(allowed_prefix):
                    valid_branch_name = True
                    break

            if not valid_branch_name:
                msg = f"Branch name '{branch}' does not start with one of {allowed_branch_prefixes}"
                violations.append(RuleViolation(self.id, msg, line_nr=1))

        return violations



# -*- coding: utf-8 -*-
# pylint: disable=too-many-function-args,unexpected-keyword-arg
from qa.shell import gitlint
from qa.base import BaseTestCase


class UserDefinedRuleTests(BaseTestCase):
    """ Integration tests for user-defined rules."""

    def test_user_defined_rules_examples1(self):
        """ Test the user defined rules in the top-level `examples/` directory """
        extra_path = self.get_example_path()
        commit_msg = "WIP: Thi$ is å title\nContent on the second line"
        self.create_simple_commit(commit_msg)
        output = gitlint("--extra-path", extra_path, _cwd=self.tmp_git_repo, _tty_in=True, _ok_code=[5])
        self.assertEqualStdout(output, self.get_expected("test_user_defined/test_user_defined_rules_examples_1"))

    def test_user_defined_rules_examples2(self):
        """ Test the user defined rules in the top-level `examples/` directory """
        extra_path = self.get_example_path()
        commit_msg = "Release: Thi$ is å title\nContent on the second line\n$This line is ignored \nThis isn't\t\n"
        self.create_simple_commit(commit_msg)
        output = gitlint("--extra-path", extra_path, _cwd=self.tmp_git_repo, _tty_in=True, _ok_code=[4])
        self.assertEqualStdout(output, self.get_expected("test_user_defined/test_user_defined_rules_examples_2"))

    def test_user_defined_rules_examples_with_config(self):
        """ Test the user defined rules in the top-level `examples/` directory """
        extra_path = self.get_example_path()
        commit_msg = "WIP: Thi$ is å title\nContent on the second line"
        self.create_simple_commit(commit_msg)
        output = gitlint("--extra-path", extra_path, "-c", "body-max-line-count.max-line-count=1",
                         _cwd=self.tmp_git_repo, _tty_in=True, _ok_code=[6])
        expected_path = "test_user_defined/test_user_defined_rules_examples_with_config_1"
        self.assertEqualStdout(output, self.get_expected(expected_path))

    def test_user_defined_rules_extra(self):
        extra_path = self.get_sample_path("user_rules/extra")
        commit_msg = "WIP: Thi$ is å title\nContent on the second line"
        self.create_simple_commit(commit_msg)
        output = gitlint("--extra-path", extra_path, _cwd=self.tmp_git_repo, _tty_in=True, _ok_code=[9])
        self.assertEqualStdout(output, self.get_expected("test_user_defined/test_user_defined_rules_extra_1",
                                                         {'repo-path': self.tmp_git_repo}))

    def test_invalid_user_defined_rules(self):
        extra_path = self.get_sample_path("user_rules/incorrect_linerule")
        self.create_simple_commit("WIP: test")
        output = gitlint("--extra-path", extra_path, _cwd=self.tmp_git_repo, _tty_in=True, _ok_code=[255])
        self.assertEqualStdout(output,
                               "Config Error: User-defined rule class 'MyUserLineRule' must have a 'validate' method\n")



# -*- coding: utf-8 -*-
from qa.shell import gitlint
from qa.base import BaseTestCase


class NamedRuleTests(BaseTestCase):
    """ Integration tests for named rules."""

    def test_named_rule(self):
        commit_msg = "WIP: thåt dûr bår\n\nSïmple commit body"
        self.create_simple_commit(commit_msg)
        config_path = self.get_sample_path("config/named-rules")
        output = gitlint("--config", config_path, _cwd=self.tmp_git_repo, _tty_in=True, _ok_code=[5])
        self.assertEqualStdout(output, self.get_expected("test_named_rules/test_named_rule_1"))

    def test_named_user_rule(self):
        commit_msg = "Normal cömmit title\n\nSïmple commit message body"
        self.create_simple_commit(commit_msg)
        config_path = self.get_sample_path("config/named-user-rules")
        extra_path = self.get_sample_path("user_rules/extra")
        output = gitlint("--extra-path", extra_path, "--config", config_path, _cwd=self.tmp_git_repo, _tty_in=True,
                         _ok_code=[9])
        self.assertEqualStdout(output, self.get_expected("test_named_rules/test_named_user_rule_1"))



# -*- coding: utf-8 -*-
from gitlint.tests.base import BaseTestCase
from gitlint import rules
from gitlint.config import LintConfig


class ConfigurationRuleTests(BaseTestCase):
    def test_ignore_by_title(self):
        commit = self.gitcommit("Releäse\n\nThis is the secōnd body line")

        # No regex specified -> Config shouldn't be changed
        rule = rules.IgnoreByTitle()
        config = LintConfig()
        rule.apply(config, commit)
        self.assertEqual(config, LintConfig())
        self.assert_logged([])  # nothing logged -> nothing ignored

        # Matching regex -> expect config to ignore all rules
        rule = rules.IgnoreByTitle({"regex": "^Releäse(.*)"})
        expected_config = LintConfig()
        expected_config.ignore = "all"
        rule.apply(config, commit)
        self.assertEqual(config, expected_config)

        expected_log_message = "DEBUG: gitlint.rules Ignoring commit because of rule 'I1': " + \
            "Commit title 'Releäse' matches the regex '^Releäse(.*)', ignoring rules: all"
        self.assert_log_contains(expected_log_message)

        # Matching regex with specific ignore
        rule = rules.IgnoreByTitle({"regex": "^Releäse(.*)",
                                    "ignore": "T1,B2"})
        expected_config = LintConfig()
        expected_config.ignore = "T1,B2"
        rule.apply(config, commit)
        self.assertEqual(config, expected_config)

        expected_log_message = "DEBUG: gitlint.rules Ignoring commit because of rule 'I1': " + \
            "Commit title 'Releäse' matches the regex '^Releäse(.*)', ignoring rules: T1,B2"

    def test_ignore_by_body(self):
        commit = self.gitcommit("Tïtle\n\nThis is\n a relëase body\n line")

        # No regex specified -> Config shouldn't be changed
        rule = rules.IgnoreByBody()
        config = LintConfig()
        rule.apply(config, commit)
        self.assertEqual(config, LintConfig())
        self.assert_logged([])  # nothing logged -> nothing ignored

        # Matching regex -> expect config to ignore all rules
        rule = rules.IgnoreByBody({"regex": "(.*)relëase(.*)"})
        expected_config = LintConfig()
        expected_config.ignore = "all"
        rule.apply(config, commit)
        self.assertEqual(config, expected_config)

        expected_log_message = "DEBUG: gitlint.rules Ignoring commit because of rule 'I2': " + \
                               "Commit message line ' a relëase body' matches the regex '(.*)relëase(.*)'," + \
                               " ignoring rules: all"
        self.assert_log_contains(expected_log_message)

        # Matching regex with specific ignore
        rule = rules.IgnoreByBody({"regex": "(.*)relëase(.*)",
                                   "ignore": "T1,B2"})
        expected_config = LintConfig()
        expected_config.ignore = "T1,B2"
        rule.apply(config, commit)
        self.assertEqual(config, expected_config)

        expected_log_message = "DEBUG: gitlint.rules Ignoring commit because of rule 'I2': " + \
            "Commit message line ' a relëase body' matches the regex '(.*)relëase(.*)', ignoring rules: T1,B2"
        self.assert_log_contains(expected_log_message)

    def test_ignore_by_author_name(self):
        commit = self.gitcommit("Tïtle\n\nThis is\n a relëase body\n line", author_name="Tëst nåme")

        # No regex specified -> Config shouldn't be changed
        rule = rules.IgnoreByAuthorName()
        config = LintConfig()
        rule.apply(config, commit)
        self.assertEqual(config, LintConfig())
        self.assert_logged([])  # nothing logged -> nothing ignored

        # Matching regex -> expect config to ignore all rules
        rule = rules.IgnoreByAuthorName({"regex": "(.*)ëst(.*)"})
        expected_config = LintConfig()
        expected_config.ignore = "all"
        rule.apply(config, commit)
        self.assertEqual(config, expected_config)

        expected_log_message = ("DEBUG: gitlint.rules Ignoring commit because of rule 'I4': "
                                "Commit Author Name 'Tëst nåme' matches the regex '(.*)ëst(.*)',"
                                " ignoring rules: all")
        self.assert_log_contains(expected_log_message)

        # Matching regex with specific ignore
        rule = rules.IgnoreByAuthorName({"regex": "(.*)nåme", "ignore": "T1,B2"})
        expected_config = LintConfig()
        expected_config.ignore = "T1,B2"
        rule.apply(config, commit)
        self.assertEqual(config, expected_config)

        expected_log_message = ("DEBUG: gitlint.rules Ignoring commit because of rule 'I4': "
                                "Commit Author Name 'Tëst nåme' matches the regex '(.*)nåme', ignoring rules: T1,B2")
        self.assert_log_contains(expected_log_message)

    def test_ignore_body_lines(self):
        commit1 = self.gitcommit("Tïtle\n\nThis is\n a relëase body\n line")
        commit2 = self.gitcommit("Tïtle\n\nThis is\n a relëase body\n line")

        # no regex specified, nothing should have happened:
        # commit and config should remain identical, log should be empty
        rule = rules.IgnoreBodyLines()
        config = LintConfig()
        rule.apply(config, commit1)
        self.assertEqual(commit1, commit2)
        self.assertEqual(config, LintConfig())
        self.assert_logged([])

        # Matching regex
        rule = rules.IgnoreBodyLines({"regex": "(.*)relëase(.*)"})
        config = LintConfig()
        rule.apply(config, commit1)
        # Our modified commit should be identical to a commit that doesn't contain the specific line
        expected_commit = self.gitcommit("Tïtle\n\nThis is\n line")
        # The original message isn't touched by this rule, this way we always have a way to reference back to it,
        # so assert it's not modified by setting it to the same as commit1
        expected_commit.message.original = commit1.message.original
        self.assertEqual(commit1, expected_commit)
        self.assertEqual(config, LintConfig())  # config shouldn't have been modified
        self.assert_log_contains("DEBUG: gitlint.rules Ignoring line ' a relëase body' because it " +
                                 "matches '(.*)relëase(.*)'")

        # Non-Matching regex: no changes expected
        commit1 = self.gitcommit("Tïtle\n\nThis is\n a relëase body\n line")
        rule = rules.IgnoreBodyLines({"regex": "(.*)föobar(.*)"})
        config = LintConfig()
        rule.apply(config, commit1)
        self.assertEqual(commit1, commit2)
        self.assertEqual(config, LintConfig())  # config shouldn't have been modified



# -*- coding: utf-8 -*-

from io import StringIO

from unittest.mock import patch  # pylint: disable=no-name-in-module, import-error

from gitlint.tests.base import BaseTestCase
from gitlint.lint import GitLinter
from gitlint.rules import RuleViolation, TitleMustNotContainWord
from gitlint.config import LintConfig, LintConfigBuilder


class LintTests(BaseTestCase):

    def test_lint_sample1(self):
        linter = GitLinter(LintConfig())
        gitcontext = self.gitcontext(self.get_sample("commit_message/sample1"))
        violations = linter.lint(gitcontext.commits[-1])
        expected_errors = [RuleViolation("T3", "Title has trailing punctuation (.)",
                                         "Commit title contåining 'WIP', as well as trailing punctuation.", 1),
                           RuleViolation("T5", "Title contains the word 'WIP' (case-insensitive)",
                                         "Commit title contåining 'WIP', as well as trailing punctuation.", 1),
                           RuleViolation("B4", "Second line is not empty", "This line should be empty", 2),
                           RuleViolation("B1", "Line exceeds max length (135>80)",
                                         "This is the first line of the commit message body and it is meant to test " +
                                         "a line that exceeds the maximum line length of 80 characters.", 3),
                           RuleViolation("B2", "Line has trailing whitespace", "This line has a tråiling space. ", 4),
                           RuleViolation("B2", "Line has trailing whitespace", "This line has a trailing tab.\t", 5),
                           RuleViolation("B3", "Line contains hard tab characters (\\t)",
                                         "This line has a trailing tab.\t", 5)]

        self.assertListEqual(violations, expected_errors)

    def test_lint_sample2(self):
        linter = GitLinter(LintConfig())
        gitcontext = self.gitcontext(self.get_sample("commit_message/sample2"))
        violations = linter.lint(gitcontext.commits[-1])
        expected = [RuleViolation("T5", "Title contains the word 'WIP' (case-insensitive)",
                                  "Just a title contåining WIP", 1),
                    RuleViolation("B6", "Body message is missing", None, 3)]

        self.assertListEqual(violations, expected)

    def test_lint_sample3(self):
        linter = GitLinter(LintConfig())
        gitcontext = self.gitcontext(self.get_sample("commit_message/sample3"))
        violations = linter.lint(gitcontext.commits[-1])

        title = " Commit title containing 'WIP', \tleading and tråiling whitespace and longer than 72 characters."
        expected = [RuleViolation("T1", "Title exceeds max length (95>72)", title, 1),
                    RuleViolation("T3", "Title has trailing punctuation (.)", title, 1),
                    RuleViolation("T4", "Title contains hard tab characters (\\t)", title, 1),
                    RuleViolation("T5", "Title contains the word 'WIP' (case-insensitive)", title, 1),
                    RuleViolation("T6", "Title has leading whitespace", title, 1),
                    RuleViolation("B4", "Second line is not empty", "This line should be empty", 2),
                    RuleViolation("B1", "Line exceeds max length (101>80)",
                                  "This is the first line is meånt to test a line that exceeds the maximum line " +
                                  "length of 80 characters.", 3),
                    RuleViolation("B2", "Line has trailing whitespace", "This line has a trailing space. ", 4),
                    RuleViolation("B2", "Line has trailing whitespace", "This line has a tråiling tab.\t", 5),
                    RuleViolation("B3", "Line contains hard tab characters (\\t)",
                                  "This line has a tråiling tab.\t", 5)]

        self.assertListEqual(violations, expected)

    def test_lint_sample4(self):
        commit = self.gitcommit(self.get_sample("commit_message/sample4"))
        config_builder = LintConfigBuilder()
        config_builder.set_config_from_commit(commit)
        linter = GitLinter(config_builder.build())
        violations = linter.lint(commit)
        # expect no violations because sample4 has a 'gitlint: disable line'
        expected = []
        self.assertListEqual(violations, expected)

    def test_lint_sample5(self):
        commit = self.gitcommit(self.get_sample("commit_message/sample5"))
        config_builder = LintConfigBuilder()
        config_builder.set_config_from_commit(commit)
        linter = GitLinter(config_builder.build())
        violations = linter.lint(commit)

        title = " Commit title containing 'WIP', \tleading and tråiling whitespace and longer than 72 characters."
        # expect only certain violations because sample5 has a 'gitlint-ignore: T3, T6, body-max-line-length'
        expected = [RuleViolation("T1", "Title exceeds max length (95>72)", title, 1),
                    RuleViolation("T4", "Title contains hard tab characters (\\t)", title, 1),
                    RuleViolation("T5", "Title contains the word 'WIP' (case-insensitive)", title, 1),
                    RuleViolation("B4", "Second line is not empty", "This line should be ëmpty", 2),
                    RuleViolation("B2", "Line has trailing whitespace", "This line has a tråiling space. ", 4),
                    RuleViolation("B2", "Line has trailing whitespace", "This line has a trailing tab.\t", 5),
                    RuleViolation("B3", "Line contains hard tab characters (\\t)",
                                  "This line has a trailing tab.\t", 5)]
        self.assertListEqual(violations, expected)

    def test_lint_meta(self):
        """ Lint sample2 but also add some metadata to the commit so we that gets linted as well """
        linter = GitLinter(LintConfig())
        gitcontext = self.gitcontext(self.get_sample("commit_message/sample2"))
        gitcontext.commits[0].author_email = "foo bår"
        violations = linter.lint(gitcontext.commits[-1])
        expected = [RuleViolation("M1", "Author email for commit is invalid", "foo bår", None),
                    RuleViolation("T5", "Title contains the word 'WIP' (case-insensitive)",
                                  "Just a title contåining WIP", 1),
                    RuleViolation("B6", "Body message is missing", None, 3)]

        self.assertListEqual(violations, expected)

    def test_lint_ignore(self):
        lint_config = LintConfig()
        lint_config.ignore = ["T1", "T3", "T4", "T5", "T6", "B1", "B2"]
        linter = GitLinter(lint_config)
        violations = linter.lint(self.gitcommit(self.get_sample("commit_message/sample3")))

        expected = [RuleViolation("B4", "Second line is not empty", "This line should be empty", 2),
                    RuleViolation("B3", "Line contains hard tab characters (\\t)",
                                  "This line has a tråiling tab.\t", 5)]

        self.assertListEqual(violations, expected)

    def test_lint_configuration_rule(self):
        # Test that all rules are ignored because of matching regex
        lint_config = LintConfig()
        lint_config.set_rule_option("I1", "regex", "^Just a title(.*)")

        linter = GitLinter(lint_config)
        violations = linter.lint(self.gitcommit(self.get_sample("commit_message/sample2")))
        self.assertListEqual(violations, [])

        # Test ignoring only certain rules
        lint_config = LintConfig()
        lint_config.set_rule_option("I1", "regex", "^Just a title(.*)")
        lint_config.set_rule_option("I1", "ignore", "B6")

        linter = GitLinter(lint_config)
        violations = linter.lint(self.gitcommit(self.get_sample("commit_message/sample2")))

        # Normally we'd expect a B6 violation, but that one is skipped because of the specific ignore set above
        expected = [RuleViolation("T5", "Title contains the word 'WIP' (case-insensitive)",
                                  "Just a title contåining WIP", 1)]

        self.assertListEqual(violations, expected)

        # Test ignoring body lines
        lint_config = LintConfig()
        linter = GitLinter(lint_config)
        lint_config.set_rule_option("I3", "regex", "(.*)tråiling(.*)")
        violations = linter.lint(self.gitcommit(self.get_sample("commit_message/sample1")))
        expected_errors = [RuleViolation("T3", "Title has trailing punctuation (.)",
                                         "Commit title contåining 'WIP', as well as trailing punctuation.", 1),
                           RuleViolation("T5", "Title contains the word 'WIP' (case-insensitive)",
                                         "Commit title contåining 'WIP', as well as trailing punctuation.", 1),
                           RuleViolation("B4", "Second line is not empty", "This line should be empty", 2),
                           RuleViolation("B1", "Line exceeds max length (135>80)",
                                         "This is the first line of the commit message body and it is meant to test " +
                                         "a line that exceeds the maximum line length of 80 characters.", 3),
                           RuleViolation("B2", "Line has trailing whitespace", "This line has a trailing tab.\t", 4),
                           RuleViolation("B3", "Line contains hard tab characters (\\t)",
                                         "This line has a trailing tab.\t", 4)]

        self.assertListEqual(violations, expected_errors)

    def test_lint_special_commit(self):
        for commit_type in ["merge", "revert", "squash", "fixup"]:
            commit = self.gitcommit(self.get_sample(f"commit_message/{commit_type}"))
            lintconfig = LintConfig()
            linter = GitLinter(lintconfig)
            violations = linter.lint(commit)
            # Even though there are a number of violations in the commit message, they are ignored because
            # we are dealing with a merge commit
            self.assertListEqual(violations, [])

            # Check that we do see violations if we disable 'ignore-merge-commits'
            setattr(lintconfig, f"ignore_{commit_type}_commits", False)
            linter = GitLinter(lintconfig)
            violations = linter.lint(commit)
            self.assertTrue(len(violations) > 0)

    def test_lint_regex_rules(self):
        """ Additional test for title-match-regex, body-match-regex """
        commit = self.gitcommit(self.get_sample("commit_message/no-violations"))
        lintconfig = LintConfig()
        linter = GitLinter(lintconfig)
        violations = linter.lint(commit)
        # No violations by default
        self.assertListEqual(violations, [])

        # Matching regexes shouldn't be a problem
        rule_regexes = [("title-match-regex", "Tïtle$"), ("body-match-regex", "Sïgned-Off-By: (.*)$")]
        for rule_regex in rule_regexes:
            lintconfig.set_rule_option(rule_regex[0], "regex", rule_regex[1])
            violations = linter.lint(commit)
            self.assertListEqual(violations, [])

        # Non-matching regexes should return violations
        rule_regexes = [("title-match-regex", ), ("body-match-regex",)]
        lintconfig.set_rule_option("title-match-regex", "regex", "^Tïtle")
        lintconfig.set_rule_option("body-match-regex", "regex", "Sügned-Off-By: (.*)$")
        expected_violations = [RuleViolation("T7", "Title does not match regex (^Tïtle)", "Normal Commit Tïtle", 1),
                               RuleViolation("B8", "Body does not match regex (Sügned-Off-By: (.*)$)", None, 6)]
        violations = linter.lint(commit)
        self.assertListEqual(violations, expected_violations)

    def test_print_violations(self):
        violations = [RuleViolation("RULE_ID_1", "Error Messåge 1", "Violating Content 1", None),
                      RuleViolation("RULE_ID_2", "Error Message 2", "Violåting Content 2", 2)]
        linter = GitLinter(LintConfig())

        # test output with increasing verbosity
        with patch('gitlint.display.stderr', new=StringIO()) as stderr:
            linter.config.verbosity = 0
            linter.print_violations(violations)
            self.assertEqual("", stderr.getvalue())

        with patch('gitlint.display.stderr', new=StringIO()) as stderr:
            linter.config.verbosity = 1
            linter.print_violations(violations)
            expected = "-: RULE_ID_1\n2: RULE_ID_2\n"
            self.assertEqual(expected, stderr.getvalue())

        with patch('gitlint.display.stderr', new=StringIO()) as stderr:
            linter.config.verbosity = 2
            linter.print_violations(violations)
            expected = "-: RULE_ID_1 Error Messåge 1\n2: RULE_ID_2 Error Message 2\n"
            self.assertEqual(expected, stderr.getvalue())

        with patch('gitlint.display.stderr', new=StringIO()) as stderr:
            linter.config.verbosity = 3
            linter.print_violations(violations)
            expected = "-: RULE_ID_1 Error Messåge 1: \"Violating Content 1\"\n" + \
                       "2: RULE_ID_2 Error Message 2: \"Violåting Content 2\"\n"
            self.assertEqual(expected, stderr.getvalue())

    def test_named_rules(self):
        """ Test that when named rules are present, both them and the original (non-named) rules executed """

        lint_config = LintConfig()
        for rule_name in ["my-ïd", "another-rule-ïd"]:
            rule_id = TitleMustNotContainWord.id + ":" + rule_name
            lint_config.rules.add_rule(TitleMustNotContainWord, rule_id)
            lint_config.set_rule_option(rule_id, "words", ["Föo"])
            linter = GitLinter(lint_config)

        violations = [RuleViolation("T5", "Title contains the word 'WIP' (case-insensitive)", "WIP: Föo bar", 1),
                      RuleViolation("T5:another-rule-ïd", "Title contains the word 'Föo' (case-insensitive)",
                                    "WIP: Föo bar", 1),
                      RuleViolation("T5:my-ïd", "Title contains the word 'Föo' (case-insensitive)",
                                    "WIP: Föo bar", 1)]
        self.assertListEqual(violations, linter.lint(self.gitcommit("WIP: Föo bar\n\nFoo bår hur dur bla bla")))

    def test_ignore_named_rules(self):
        """ Test that named rules can be ignored """

        # Add named rule to lint config
        config_builder = LintConfigBuilder()
        rule_id = TitleMustNotContainWord.id + ":my-ïd"
        config_builder.set_option(rule_id, "words", ["Föo"])
        lint_config = config_builder.build()
        linter = GitLinter(lint_config)
        commit = self.gitcommit("WIP: Föo bar\n\nFoo bår hur dur bla bla")

        # By default, we expect both the violations of the regular rule as well as the named rule to show up
        violations = [RuleViolation("T5", "Title contains the word 'WIP' (case-insensitive)", "WIP: Föo bar", 1),
                      RuleViolation("T5:my-ïd", "Title contains the word 'Föo' (case-insensitive)",
                                    "WIP: Föo bar", 1)]
        self.assertListEqual(violations, linter.lint(commit))

        # ignore regular rule: only named rule violations show up
        lint_config.ignore = ["T5"]
        self.assertListEqual(violations[1:], linter.lint(commit))

        # ignore named rule by id: only regular rule violations show up
        lint_config.ignore = [rule_id]
        self.assertListEqual(violations[:-1], linter.lint(commit))

        # ignore named rule by name: only regular rule violations show up
        lint_config.ignore = [TitleMustNotContainWord.name + ":my-ïd"]
        self.assertListEqual(violations[:-1], linter.lint(commit))



# -*- coding: utf-8 -*-
from gitlint.tests.base import BaseTestCase
from gitlint.cache import PropertyCache, cache


class CacheTests(BaseTestCase):

    class MyClass(PropertyCache):
        """ Simple class that has cached properties, used for testing. """

        def __init__(self):
            PropertyCache.__init__(self)
            self.counter = 0

        @property
        @cache
        def foo(self):
            self.counter += 1
            return "bår"

        @property
        @cache(cachekey="hür")
        def bar(self):
            self.counter += 1
            return "fōo"

    def test_cache(self):
        # Init new class with cached properties
        myclass = self.MyClass()
        self.assertEqual(myclass.counter, 0)
        self.assertDictEqual(myclass._cache, {})

        # Assert that function is called on first access, cache is set
        self.assertEqual(myclass.foo, "bår")
        self.assertEqual(myclass.counter, 1)
        self.assertDictEqual(myclass._cache, {"foo": "bår"})

        # After function is not called on subsequent access, cache is still set
        self.assertEqual(myclass.foo, "bår")
        self.assertEqual(myclass.counter, 1)
        self.assertDictEqual(myclass._cache, {"foo": "bår"})

    def test_cache_custom_key(self):
        # Init new class with cached properties
        myclass = self.MyClass()
        self.assertEqual(myclass.counter, 0)
        self.assertDictEqual(myclass._cache, {})

        # Assert that function is called on first access, cache is set with custom key
        self.assertEqual(myclass.bar, "fōo")
        self.assertEqual(myclass.counter, 1)
        self.assertDictEqual(myclass._cache, {"hür": "fōo"})

        # After function is not called on subsequent access, cache is still set
        self.assertEqual(myclass.bar, "fōo")
        self.assertEqual(myclass.counter, 1)
        self.assertDictEqual(myclass._cache, {"hür": "fōo"})



# -*- coding: utf-8 -*-
from gitlint.tests.base import BaseTestCase
from gitlint import rules


class BodyRuleTests(BaseTestCase):
    def test_max_line_length(self):
        rule = rules.BodyMaxLineLength()

        # assert no error
        violation = rule.validate("å" * 80, None)
        self.assertIsNone(violation)

        # assert error on line length > 80
        expected_violation = rules.RuleViolation("B1", "Line exceeds max length (81>80)", "å" * 81)
        violations = rule.validate("å" * 81, None)
        self.assertListEqual(violations, [expected_violation])

        # set line length to 120, and check no violation on length 73
        rule = rules.BodyMaxLineLength({'line-length': 120})
        violations = rule.validate("å" * 73, None)
        self.assertIsNone(violations)

        # assert raise on 121
        expected_violation = rules.RuleViolation("B1", "Line exceeds max length (121>120)", "å" * 121)
        violations = rule.validate("å" * 121, None)
        self.assertListEqual(violations, [expected_violation])

    def test_trailing_whitespace(self):
        rule = rules.BodyTrailingWhitespace()

        # assert no error
        violations = rule.validate("å", None)
        self.assertIsNone(violations)

        # trailing space
        expected_violation = rules.RuleViolation("B2", "Line has trailing whitespace", "å ")
        violations = rule.validate("å ", None)
        self.assertListEqual(violations, [expected_violation])

        # trailing tab
        expected_violation = rules.RuleViolation("B2", "Line has trailing whitespace", "å\t")
        violations = rule.validate("å\t", None)
        self.assertListEqual(violations, [expected_violation])

    def test_hard_tabs(self):
        rule = rules.BodyHardTab()

        # assert no error
        violations = rule.validate("This is ã test", None)
        self.assertIsNone(violations)

        # contains hard tab
        expected_violation = rules.RuleViolation("B3", "Line contains hard tab characters (\\t)", "This is å\ttest")
        violations = rule.validate("This is å\ttest", None)
        self.assertListEqual(violations, [expected_violation])

    def test_body_first_line_empty(self):
        rule = rules.BodyFirstLineEmpty()

        # assert no error
        commit = self.gitcommit("Tïtle\n\nThis is the secōnd body line")
        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # second line not empty
        expected_violation = rules.RuleViolation("B4", "Second line is not empty", "nöt empty", 2)

        commit = self.gitcommit("Tïtle\nnöt empty\nThis is the secönd body line")
        violations = rule.validate(commit)
        self.assertListEqual(violations, [expected_violation])

    def test_body_min_length(self):
        rule = rules.BodyMinLength()

        # assert no error - body is long enough
        commit = self.gitcommit("Title\n\nThis is the second body line\n")

        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # assert no error - no body
        commit = self.gitcommit("Tïtle\n")
        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # body is too short
        expected_violation = rules.RuleViolation("B5", "Body message is too short (8<20)", "töoshort", 3)

        commit = self.gitcommit("Tïtle\n\ntöoshort\n")
        violations = rule.validate(commit)
        self.assertListEqual(violations, [expected_violation])

        # assert error - short across multiple lines
        expected_violation = rules.RuleViolation("B5", "Body message is too short (11<20)", "secöndthïrd", 3)
        commit = self.gitcommit("Tïtle\n\nsecönd\nthïrd\n")
        violations = rule.validate(commit)
        self.assertListEqual(violations, [expected_violation])

        # set line length to 120, and check violation on length 21
        expected_violation = rules.RuleViolation("B5", "Body message is too short (21<120)", "å" * 21, 3)

        rule = rules.BodyMinLength({'min-length': 120})
        commit = self.gitcommit("Title\n\n{0}\n".format("å" * 21))  # pylint: disable=consider-using-f-string
        violations = rule.validate(commit)
        self.assertListEqual(violations, [expected_violation])

        # Make sure we don't get the error if the body-length is exactly the min-length
        rule = rules.BodyMinLength({'min-length': 8})
        commit = self.gitcommit("Tïtle\n\n{0}\n".format("å" * 8))  # pylint: disable=consider-using-f-string
        violations = rule.validate(commit)
        self.assertIsNone(violations)

    def test_body_missing(self):
        rule = rules.BodyMissing()

        # assert no error - body is present
        commit = self.gitcommit("Tïtle\n\nThis ïs the first body line\n")
        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # body is too short
        expected_violation = rules.RuleViolation("B6", "Body message is missing", None, 3)

        commit = self.gitcommit("Tïtle\n")
        violations = rule.validate(commit)
        self.assertListEqual(violations, [expected_violation])

    def test_body_missing_multiple_empty_new_lines(self):
        rule = rules.BodyMissing()

        # body is too short
        expected_violation = rules.RuleViolation("B6", "Body message is missing", None, 3)

        commit = self.gitcommit("Tïtle\n\n\n\n")
        violations = rule.validate(commit)
        self.assertListEqual(violations, [expected_violation])

    def test_body_missing_merge_commit(self):
        rule = rules.BodyMissing()

        # assert no error - merge commit
        commit = self.gitcommit("Merge: Tïtle\n")
        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # assert error for merge commits if ignore-merge-commits is disabled
        rule = rules.BodyMissing({'ignore-merge-commits': False})
        violations = rule.validate(commit)
        expected_violation = rules.RuleViolation("B6", "Body message is missing", None, 3)
        self.assertListEqual(violations, [expected_violation])

    def test_body_changed_file_mention(self):
        rule = rules.BodyChangedFileMention()

        # assert no error when no files have changed and no files need to be mentioned
        commit = self.gitcommit("This is a test\n\nHere is a mention of föo/test.py")
        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # assert no error when no files have changed but certain files need to be mentioned on change
        rule = rules.BodyChangedFileMention({'files': "bar.txt,föo/test.py"})
        commit = self.gitcommit("This is a test\n\nHere is a mention of föo/test.py")
        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # assert no error if a file has changed and is mentioned
        commit = self.gitcommit("This is a test\n\nHere is a mention of föo/test.py", ["föo/test.py"])
        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # assert no error if multiple files have changed and are mentioned
        commit_msg = "This is a test\n\nHere is a mention of föo/test.py\nAnd here is a mention of bar.txt"
        commit = self.gitcommit(commit_msg, ["föo/test.py", "bar.txt"])
        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # assert error if file has changed and is not mentioned
        commit_msg = "This is a test\n\nHere is å mention of\nAnd here is a mention of bar.txt"
        commit = self.gitcommit(commit_msg, ["föo/test.py", "bar.txt"])
        violations = rule.validate(commit)
        expected_violation = rules.RuleViolation("B7", "Body does not mention changed file 'föo/test.py'", None, 4)
        self.assertEqual([expected_violation], violations)

        # assert multiple errors if multiple files have changed and are not mentioned
        commit_msg = "This is å test\n\nHere is a mention of\nAnd here is a mention of"
        commit = self.gitcommit(commit_msg, ["föo/test.py", "bar.txt"])
        violations = rule.validate(commit)
        expected_violation_2 = rules.RuleViolation("B7", "Body does not mention changed file 'bar.txt'", None, 4)
        self.assertEqual([expected_violation_2, expected_violation], violations)

    def test_body_match_regex(self):
        # We intentionally add 2 newlines at the end of our commit message as that's how git will pass the
        # message. This way we also test that the rule strips off the last line.
        commit = self.gitcommit("US1234: åbc\nIgnored\nBödy\nFöo\nMy-Commit-Tag: föo\n\n")

        # assert no violation on default regex (=everything allowed)
        rule = rules.BodyRegexMatches()
        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # assert no violation on matching regex
        # (also note that first body line - in between title and rest of body - is ignored)
        rule = rules.BodyRegexMatches({'regex': "^Bödy(.*)"})
        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # assert we can do end matching (and last empty line is ignored)
        # (also note that first body line - in between title and rest of body - is ignored)
        rule = rules.BodyRegexMatches({'regex': "My-Commit-Tag: föo$"})
        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # common use-case: matching that a given line is present
        rule = rules.BodyRegexMatches({'regex': "(.*)Föo(.*)"})
        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # assert violation on non-matching body
        rule = rules.BodyRegexMatches({'regex': "^Tëst(.*)Foo"})
        violations = rule.validate(commit)
        expected_violation = rules.RuleViolation("B8", "Body does not match regex (^Tëst(.*)Foo)", None, 6)
        self.assertListEqual(violations, [expected_violation])

        # assert no violation on None regex
        rule = rules.BodyRegexMatches({'regex': None})
        violations = rule.validate(commit)
        self.assertIsNone(violations)

        # Assert no issues when there's no body or a weird body variation
        bodies = ["åbc", "åbc\n", "åbc\nföo\n", "åbc\n\n", "åbc\nföo\nblå", "åbc\nföo\nblå\n"]
        for body in bodies:
            commit = self.gitcommit(body)
            rule = rules.BodyRegexMatches({'regex': ".*"})
            violations = rule.validate(commit)
            self.assertIsNone(violations)



import logging
import os

import arrow

from gitlint import shell as sh
# import exceptions separately, this makes it a little easier to mock them out in the unit tests
from gitlint.shell import CommandNotFound, ErrorReturnCode

from gitlint.cache import PropertyCache, cache
from gitlint.exception import GitlintError

# For now, the git date format we use is fixed, but technically this format is determined by `git config log.date`
# We should fix this at some point :-)
GIT_TIMEFORMAT = "YYYY-MM-DD HH:mm:ss Z"

LOG = logging.getLogger(__name__)


class GitContextError(GitlintError):
    """ Exception indicating there is an issue with the git context """
    pass


class GitNotInstalledError(GitContextError):
    def __init__(self):
        super().__init__(
            "'git' command not found. You need to install git to use gitlint on a local repository. " +
            "See https://git-scm.com/book/en/v2/Getting-Started-Installing-Git on how to install git.")


class GitExitCodeError(GitContextError):
    def __init__(self, command, stderr):
        self.command = command
        self.stderr = stderr
        super().__init__(f"An error occurred while executing '{command}': {stderr}")


def _git(*command_parts, **kwargs):
    """ Convenience function for running git commands. Automatically deals with exceptions and unicode. """
    git_kwargs = {'_tty_out': False}
    git_kwargs.update(kwargs)
    try:
        LOG.debug(command_parts)
        result = sh.git(*command_parts, **git_kwargs)  # pylint: disable=unexpected-keyword-arg
        # If we reach this point and the result has an exit_code that is larger than 0, this means that we didn't
        # get an exception (which is the default sh behavior for non-zero exit codes) and so the user is expecting
        # a non-zero exit code -> just return the entire result
        if hasattr(result, 'exit_code') and result.exit_code > 0:
            return result
        return str(result)
    except CommandNotFound as e:
        raise GitNotInstalledError from e
    except ErrorReturnCode as e:  # Something went wrong while executing the git command
        error_msg = e.stderr.strip()
        error_msg_lower = error_msg.lower()
        if '_cwd' in git_kwargs and b"not a git repository" in error_msg_lower:
            raise GitContextError(f"{git_kwargs['_cwd']} is not a git repository.") from e

        if (b"does not have any commits yet" in error_msg_lower or
                b"ambiguous argument 'head': unknown revision" in error_msg_lower):
            msg = "Current branch has no commits. Gitlint requires at least one commit to function."
            raise GitContextError(msg) from e

        raise GitExitCodeError(e.full_cmd, error_msg) from e


def git_version():
    """ Determine the git version installed on this host by calling git --version"""
    return _git("--version").replace("\n", "")


def git_commentchar(repository_path=None):
    """ Shortcut for retrieving comment char from git config """
    commentchar = _git("config", "--get", "core.commentchar", _cwd=repository_path, _ok_code=[0, 1])
    # git will return an exit code of 1 if it can't find a config value, in this case we fall-back to # as commentchar
    if hasattr(commentchar, 'exit_code') and commentchar.exit_code == 1:  # pylint: disable=no-member
        commentchar = "#"
    return commentchar.replace("\n", "")


def git_hooks_dir(repository_path):
    """ Determine hooks directory for a given target dir """
    hooks_dir = _git("rev-parse", "--git-path", "hooks", _cwd=repository_path)
    hooks_dir = hooks_dir.replace("\n", "")
    return os.path.realpath(os.path.join(repository_path, hooks_dir))


class GitCommitMessage:
    """ Class representing a git commit message. A commit message consists of the following:
      - context: The `GitContext` this commit message is part of
      - original: The actual commit message as returned by `git log`
      - full: original, but stripped of any comments
      - title: the first line of full
      - body: all lines following the title
    """
    def __init__(self, context, original=None, full=None, title=None, body=None):
        self.context = context
        self.original = original
        self.full = full
        self.title = title
        self.body = body

    @staticmethod
    def from_full_message(context, commit_msg_str):
        """  Parses a full git commit message by parsing a given string into the different parts of a commit message """
        all_lines = commit_msg_str.splitlines()
        cutline = f"{context.commentchar} ------------------------ >8 ------------------------"
        try:
            cutline_index = all_lines.index(cutline)
        except ValueError:
            cutline_index = None
        lines = [line for line in all_lines[:cutline_index] if not line.startswith(context.commentchar)]
        full = "\n".join(lines)
        title = lines[0] if lines else ""
        body = lines[1:] if len(lines) > 1 else []
        return GitCommitMessage(context=context, original=commit_msg_str, full=full, title=title, body=body)

    def __str__(self):
        return self.full

    def __eq__(self, other):
        return (isinstance(other, GitCommitMessage) and self.original == other.original
                and self.full == other.full and self.title == other.title and self.body == other.body)  # noqa


class GitCommit:
    """ Class representing a git commit.
        A commit consists of: context, message, author name, author email, date, list of parent commit shas,
        list of changed files, list of branch names.
        In the context of gitlint, only the git context and commit message are required.
    """

    def __init__(self, context, message, sha=None, date=None, author_name=None,  # pylint: disable=too-many-arguments
                 author_email=None, parents=None, changed_files=None, branches=None):
        self.context = context
        self.message = message
        self.sha = sha
        self.date = date
        self.author_name = author_name
        self.author_email = author_email
        self.parents = parents or []  # parent commit hashes
        self.changed_files = changed_files or []
        self.branches = branches or []

    @property
    def is_merge_commit(self):
        return self.message.title.startswith("Merge")

    @property
    def is_fixup_commit(self):
        return self.message.title.startswith("fixup!")

    @property
    def is_squash_commit(self):
        return self.message.title.startswith("squash!")

    @property
    def is_revert_commit(self):
        return self.message.title.startswith("Revert")

    def __str__(self):
        date_str = arrow.get(self.date).format(GIT_TIMEFORMAT) if self.date else None
        return (f"--- Commit Message ----\n{self.message}\n"
                "--- Meta info ---------\n"
                f"Author: {self.author_name} <{self.author_email}>\n"
                f"Date:   {date_str}\n"
                f"is-merge-commit:  {self.is_merge_commit}\n"
                f"is-fixup-commit:  {self.is_fixup_commit}\n"
                f"is-squash-commit: {self.is_squash_commit}\n"
                f"is-revert-commit: {self.is_revert_commit}\n"
                f"Branches: {self.branches}\n"
                f"Changed Files: {self.changed_files}\n"
                "-----------------------")

    def __eq__(self, other):
        # skip checking the context as context refers back to this obj, this will trigger a cyclic dependency
        return (isinstance(other, GitCommit) and self.message == other.message
                and self.sha == other.sha and self.author_name == other.author_name
                and self.author_email == other.author_email
                and self.date == other.date and self.parents == other.parents
                and self.is_merge_commit == other.is_merge_commit and self.is_fixup_commit == other.is_fixup_commit
                and self.is_squash_commit == other.is_squash_commit and self.is_revert_commit == other.is_revert_commit
                and self.changed_files == other.changed_files and self.branches == other.branches) # noqa


class LocalGitCommit(GitCommit, PropertyCache):
    """ Class representing a git commit that exists in the local git repository.
        This class uses lazy loading: it defers reading information from the local git repository until the associated
        property is accessed for the first time. Properties are then cached for subsequent access.

        This approach ensures that we don't do 'expensive' git calls when certain properties are not actually used.
        In addition, reading the required info when it's needed rather than up front avoids adding delay during gitlint
        startup time and reduces gitlint's memory footprint.
     """
    def __init__(self, context, sha):  # pylint: disable=super-init-not-called
        PropertyCache.__init__(self)
        self.context = context
        self.sha = sha

    def _log(self):
        """ Does a call to `git log` to determine a bunch of information about the commit. """
        long_format = "--pretty=%aN%x00%aE%x00%ai%x00%P%n%B"
        raw_commit = _git("log", self.sha, "-1", long_format, _cwd=self.context.repository_path).split("\n")

        (name, email, date, parents), commit_msg = raw_commit[0].split('\x00'), "\n".join(raw_commit[1:])

        commit_parents = parents.split(" ")
        commit_is_merge_commit = len(commit_parents) > 1

        # "YYYY-MM-DD HH:mm:ss Z" -> ISO 8601-like format
        # Use arrow for datetime parsing, because apparently python is quirky around ISO-8601 dates:
        # http://stackoverflow.com/a/30696682/381010
        commit_date = arrow.get(date, GIT_TIMEFORMAT).datetime

        # Create Git commit object with the retrieved info
        commit_msg_obj = GitCommitMessage.from_full_message(self.context, commit_msg)

        self._cache.update({'message': commit_msg_obj, 'author_name': name, 'author_email': email, 'date': commit_date,
                            'parents': commit_parents, 'is_merge_commit': commit_is_merge_commit})

    @property
    def message(self):
        return self._try_cache("message", self._log)

    @property
    def author_name(self):
        return self._try_cache("author_name", self._log)

    @property
    def author_email(self):
        return self._try_cache("author_email", self._log)

    @property
    def date(self):
        return self._try_cache("date", self._log)

    @property
    def parents(self):
        return self._try_cache("parents", self._log)

    @property
    def branches(self):
        def cache_branches():
            # We have to parse 'git branch --contains <sha>' instead of 'git for-each-ref' to be compatible with
            # git versions < 2.7.0
            # https://stackoverflow.com/questions/45173979/can-i-force-git-branch-contains-tag-to-not-print-the-asterisk
            branches = _git("branch", "--contains", self.sha, _cwd=self.context.repository_path).split("\n")

            # This means that we need to remove any leading * that indicates the current branch. Note that we can
            # safely do this since git branches cannot contain '*' anywhere, so if we find an '*' we know it's output
            # from the git CLI and not part of the branch name. See https://git-scm.com/docs/git-check-ref-format
            # We also drop the last empty line from the output.
            self._cache['branches'] = [branch.replace("*", "").strip() for branch in branches[:-1]]

        return self._try_cache("branches", cache_branches)

    @property
    def is_merge_commit(self):
        return self._try_cache("is_merge_commit", self._log)

    @property
    def changed_files(self):
        def cache_changed_files():
            self._cache['changed_files'] = _git("diff-tree", "--no-commit-id", "--name-only", "-r", "--root",
                                                self.sha, _cwd=self.context.repository_path).split()

        return self._try_cache("changed_files", cache_changed_files)


class StagedLocalGitCommit(GitCommit, PropertyCache):
    """ Class representing a git commit that has been staged, but not committed.

        Other than the commit message itself (and changed files), a lot of information is actually not known at staging
        time, since the commit hasn't happened yet. However, we can make educated guesses based on existing repository
        information.
    """

    def __init__(self, context, commit_message):  # pylint: disable=super-init-not-called
        PropertyCache.__init__(self)
        self.context = context
        self.message = commit_message
        self.sha = None
        self.parents = []  # Not really possible to determine before a commit

    @property
    @cache
    def author_name(self):
        try:
            return _git("config", "--get", "user.name", _cwd=self.context.repository_path).strip()
        except GitExitCodeError as e:
            raise GitContextError("Missing git configuration: please set user.name") from e

    @property
    @cache
    def author_email(self):
        try:
            return _git("config", "--get", "user.email", _cwd=self.context.repository_path).strip()
        except GitExitCodeError as e:
            raise GitContextError("Missing git configuration: please set user.email") from e

    @property
    @cache
    def date(self):
        # We don't know the actual commit date yet, but we make a pragmatic trade-off here by providing the current date
        # We get current date from arrow, reformat in git date format, then re-interpret it as a date.
        # This ensure we capture the same precision and timezone information that git does.
        return arrow.get(arrow.now().format(GIT_TIMEFORMAT), GIT_TIMEFORMAT).datetime

    @property
    @cache
    def branches(self):
        # We don't know the branch this commit will be part of yet, but we're pragmatic here and just return the
        # current branch, as for all intents and purposes, this will be what the user is looking for.
        return [self.context.current_branch]

    @property
    def changed_files(self):
        return _git("diff", "--staged", "--name-only", "-r", _cwd=self.context.repository_path).split()


class GitContext(PropertyCache):
    """ Class representing the git context in which gitlint is operating: a data object storing information about
    the git repository that gitlint is linting.
    """

    def __init__(self, repository_path=None):
        PropertyCache.__init__(self)
        self.commits = []
        self.repository_path = repository_path

    @property
    @cache
    def commentchar(self):
        return git_commentchar(self.repository_path)

    @property
    @cache
    def current_branch(self):
        current_branch = _git("rev-parse", "--abbrev-ref", "HEAD", _cwd=self.repository_path).strip()
        return current_branch

    @staticmethod
    def from_commit_msg(commit_msg_str):
        """ Determines git context based on a commit message.
        :param commit_msg_str: Full git commit message.
        """
        context = GitContext()
        commit_msg_obj = GitCommitMessage.from_full_message(context, commit_msg_str)
        commit = GitCommit(context, commit_msg_obj)
        context.commits.append(commit)
        return context

    @staticmethod
    def from_staged_commit(commit_msg_str, repository_path):
        """ Determines git context based on a commit message that is a staged commit for a local git repository.
        :param commit_msg_str: Full git commit message.
        :param repository_path: Path to the git repository to retrieve the context from
        """
        context = GitContext(repository_path=repository_path)
        commit_msg_obj = GitCommitMessage.from_full_message(context, commit_msg_str)
        commit = StagedLocalGitCommit(context, commit_msg_obj)
        context.commits.append(commit)
        return context

    @staticmethod
    def from_local_repository(repository_path, refspec=None, commit_hash=None):
        """ Retrieves the git context from a local git repository.
        :param repository_path: Path to the git repository to retrieve the context from
        :param refspec: The commit(s) to retrieve (mutually exclusive with `commit_hash`)
        :param commit_hash: Hash of the commit to retrieve (mutually exclusive with `refspec`)
        """

        context = GitContext(repository_path=repository_path)

        if refspec:
            sha_list = _git("rev-list", refspec, _cwd=repository_path).split()
        elif commit_hash:  # Single commit, just pass it to `git log -1`
            # Even though we have already been passed the commit hash, we ask git to retrieve this hash and
            # return it to us. This way we verify that the passed hash is a valid hash for the target repo and we
            # also convert it to the full hash format (we might have been passed a short hash).
            sha_list = [_git("log", "-1", commit_hash, "--pretty=%H", _cwd=repository_path).replace("\n", "")]
        else:  # If no refspec is defined, fallback to the last commit on the current branch
            # We tried many things here e.g.: defaulting to e.g. HEAD or HEAD^... (incl. dealing with
            # repos that only have a single commit - HEAD^... doesn't work there), but then we still get into
            # problems with e.g. merge commits. Easiest solution is just taking the SHA from `git log -1`.
            sha_list = [_git("log", "-1", "--pretty=%H", _cwd=repository_path).replace("\n", "")]

        for sha in sha_list:
            commit = LocalGitCommit(context, sha)
            context.commits.append(commit)

        return context

    def __eq__(self, other):
        return (isinstance(other, GitContext) and self.commits == other.commits
                and self.repository_path == other.repository_path
                and self.commentchar == other.commentchar and self.current_branch == other.current_branch)  # noqa



# -*- coding: utf-8 -*-

from unittest.mock import patch

from gitlint import rules
from gitlint.config import LintConfig, LintConfigError, LintConfigGenerator, GITLINT_CONFIG_TEMPLATE_SRC_PATH
from gitlint import options
from gitlint.tests.base import BaseTestCase


class LintConfigTests(BaseTestCase):

    def test_set_rule_option(self):
        config = LintConfig()

        # assert default title line-length
        self.assertEqual(config.get_rule_option('title-max-length', 'line-length'), 72)

        # change line length and assert it is set
        config.set_rule_option('title-max-length', 'line-length', 60)
        self.assertEqual(config.get_rule_option('title-max-length', 'line-length'), 60)

    def test_set_rule_option_negative(self):
        config = LintConfig()

        # non-existing rule
        expected_error_msg = "No such rule 'föobar'"
        with self.assertRaisesMessage(LintConfigError, expected_error_msg):
            config.set_rule_option(u'föobar', u'lïne-length', 60)

        # non-existing option
        expected_error_msg = "Rule 'title-max-length' has no option 'föobar'"
        with self.assertRaisesMessage(LintConfigError, expected_error_msg):
            config.set_rule_option('title-max-length', u'föobar', 60)

        # invalid option value
        expected_error_msg = "'föo' is not a valid value for option 'title-max-length.line-length'. " + \
                             "Option 'line-length' must be a positive integer (current value: 'föo')."
        with self.assertRaisesMessage(LintConfigError, expected_error_msg):
            config.set_rule_option('title-max-length', 'line-length', "föo")

    def test_set_general_option(self):
        config = LintConfig()

        # Check that default general options are correct
        self.assertTrue(config.ignore_merge_commits)
        self.assertTrue(config.ignore_fixup_commits)
        self.assertTrue(config.ignore_squash_commits)
        self.assertTrue(config.ignore_revert_commits)

        self.assertFalse(config.ignore_stdin)
        self.assertFalse(config.staged)
        self.assertFalse(config.fail_without_commits)
        self.assertFalse(config.debug)
        self.assertEqual(config.verbosity, 3)
        active_rule_classes = tuple(type(rule) for rule in config.rules)
        self.assertTupleEqual(active_rule_classes, config.default_rule_classes)

        # ignore - set by string
        config.set_general_option("ignore", "title-trailing-whitespace, B2")
        self.assertEqual(config.ignore, ["title-trailing-whitespace", "B2"])

        # ignore - set by list
        config.set_general_option("ignore", ["T1", "B3"])
        self.assertEqual(config.ignore, ["T1", "B3"])

        # verbosity
        config.set_general_option("verbosity", 1)
        self.assertEqual(config.verbosity, 1)

        # ignore_merge_commit
        config.set_general_option("ignore-merge-commits", "false")
        self.assertFalse(config.ignore_merge_commits)

        # ignore_fixup_commit
        config.set_general_option("ignore-fixup-commits", "false")
        self.assertFalse(config.ignore_fixup_commits)

        # ignore_squash_commit
        config.set_general_option("ignore-squash-commits", "false")
        self.assertFalse(config.ignore_squash_commits)

        # ignore_revert_commit
        config.set_general_option("ignore-revert-commits", "false")
        self.assertFalse(config.ignore_revert_commits)

        # debug
        config.set_general_option("debug", "true")
        self.assertTrue(config.debug)

        # ignore-stdin
        config.set_general_option("ignore-stdin", "true")
        self.assertTrue(config.debug)

        # staged
        config.set_general_option("staged", "true")
        self.assertTrue(config.staged)

        # fail-without-commits
        config.set_general_option("fail-without-commits", "true")
        self.assertTrue(config.fail_without_commits)

        # target
        config.set_general_option("target", self.SAMPLES_DIR)
        self.assertEqual(config.target, self.SAMPLES_DIR)

        # extra_path has its own test: test_extra_path and test_extra_path_negative
        # contrib has its own tests: test_contrib and test_contrib_negative

    def test_contrib(self):
        config = LintConfig()
        contrib_rules = ["contrib-title-conventional-commits", "CC1"]
        config.set_general_option("contrib", ",".join(contrib_rules))
        self.assertEqual(config.contrib, contrib_rules)

        # Check contrib-title-conventional-commits contrib rule
        actual_rule = config.rules.find_rule("contrib-title-conventional-commits")
        self.assertTrue(actual_rule.is_contrib)

        self.assertEqual(str(type(actual_rule)), "<class 'conventional_commit.ConventionalCommit'>")
        self.assertEqual(actual_rule.id, 'CT1')
        self.assertEqual(actual_rule.name, u'contrib-title-conventional-commits')
        self.assertEqual(actual_rule.target, rules.CommitMessageTitle)

        expected_rule_option = options.ListOption(
            "types",
            ["fix", "feat", "chore", "docs", "style", "refactor", "perf", "test", "revert", "ci", "build"],
            "Comma separated list of allowed commit types.",
        )

        self.assertListEqual(actual_rule.options_spec, [expected_rule_option])
        self.assertDictEqual(actual_rule.options, {'types': expected_rule_option})

        # Check contrib-body-requires-signed-off-by contrib rule
        actual_rule = config.rules.find_rule("contrib-body-requires-signed-off-by")
        self.assertTrue(actual_rule.is_contrib)

        self.assertEqual(str(type(actual_rule)), "<class 'signedoff_by.SignedOffBy'>")
        self.assertEqual(actual_rule.id, 'CC1')
        self.assertEqual(actual_rule.name, u'contrib-body-requires-signed-off-by')

        # reset value (this is a different code path)
        config.set_general_option("contrib", "contrib-body-requires-signed-off-by")
        self.assertEqual(actual_rule, config.rules.find_rule("contrib-body-requires-signed-off-by"))
        self.assertIsNone(config.rules.find_rule("contrib-title-conventional-commits"))

        # empty value
        config.set_general_option("contrib", "")
        self.assertListEqual(config.contrib, [])

    def test_contrib_negative(self):
        config = LintConfig()
        # non-existent contrib rule
        with self.assertRaisesMessage(LintConfigError, "No contrib rule with id or name 'föo' found."):
            config.contrib = "contrib-title-conventional-commits,föo"

        # UserRuleError, RuleOptionError should be re-raised as LintConfigErrors
        side_effects = [rules.UserRuleError("üser-rule"), options.RuleOptionError("rüle-option")]
        for side_effect in side_effects:
            with patch('gitlint.config.rule_finder.find_rule_classes', side_effect=side_effect):
                with self.assertRaisesMessage(LintConfigError, str(side_effect)):
                    config.contrib = "contrib-title-conventional-commits"

    def test_extra_path(self):
        config = LintConfig()

        config.set_general_option("extra-path", self.get_user_rules_path())
        self.assertEqual(config.extra_path, self.get_user_rules_path())
        actual_rule = config.rules.find_rule('UC1')
        self.assertTrue(actual_rule.is_user_defined)
        self.assertEqual(str(type(actual_rule)), "<class 'my_commit_rules.MyUserCommitRule'>")
        self.assertEqual(actual_rule.id, 'UC1')
        self.assertEqual(actual_rule.name, u'my-üser-commit-rule')
        self.assertEqual(actual_rule.target, None)
        expected_rule_option = options.IntOption('violation-count', 1, "Number of violåtions to return")
        self.assertListEqual(actual_rule.options_spec, [expected_rule_option])
        self.assertDictEqual(actual_rule.options, {'violation-count': expected_rule_option})

        # reset value (this is a different code path)
        config.set_general_option("extra-path", self.SAMPLES_DIR)
        self.assertEqual(config.extra_path, self.SAMPLES_DIR)
        self.assertIsNone(config.rules.find_rule("UC1"))

    def test_extra_path_negative(self):
        config = LintConfig()
        regex = "Option extra-path must be either an existing directory or file (current value: 'föo/bar')"
        # incorrect extra_path
        with self.assertRaisesMessage(LintConfigError, regex):
            config.extra_path = "föo/bar"

        # extra path contains classes with errors
        with self.assertRaisesMessage(LintConfigError,
                                      "User-defined rule class 'MyUserLineRule' must have a 'validate' method"):
            config.extra_path = self.get_sample_path("user_rules/incorrect_linerule")

    def test_set_general_option_negative(self):
        config = LintConfig()

        # Note that we shouldn't test whether we can set unicode because python just doesn't allow unicode attributes
        with self.assertRaisesMessage(LintConfigError, "'foo' is not a valid gitlint option"):
            config.set_general_option("foo", "bår")

        # try setting _config_path, this is a real attribute of LintConfig, but the code should prevent it from
        # being set
        with self.assertRaisesMessage(LintConfigError, "'_config_path' is not a valid gitlint option"):
            config.set_general_option("_config_path", "bår")

        # invalid verbosity
        incorrect_values = [-1, "föo"]
        for value in incorrect_values:
            expected_msg = f"Option 'verbosity' must be a positive integer (current value: '{value}')"
            with self.assertRaisesMessage(LintConfigError, expected_msg):
                config.verbosity = value

        incorrect_values = [4]
        for value in incorrect_values:
            with self.assertRaisesMessage(LintConfigError, "Option 'verbosity' must be set between 0 and 3"):
                config.verbosity = value

        # invalid ignore_xxx_commits
        ignore_attributes = ["ignore_merge_commits", "ignore_fixup_commits", "ignore_squash_commits",
                             "ignore_revert_commits"]
        incorrect_values = [-1, 4, "föo"]
        for attribute in ignore_attributes:
            for value in incorrect_values:
                option_name = attribute.replace("_", "-")
                with self.assertRaisesMessage(LintConfigError,
                                              f"Option '{option_name}' must be either 'true' or 'false'"):
                    setattr(config, attribute, value)

        # invalid ignore -> not here because ignore is a ListOption which converts everything to a string before
        # splitting which means it it will accept just about everything

        # invalid boolean options
        for attribute in ['debug', 'staged', 'ignore_stdin', 'fail_without_commits']:
            option_name = attribute.replace("_", "-")
            with self.assertRaisesMessage(LintConfigError,
                                          f"Option '{option_name}' must be either 'true' or 'false'"):
                setattr(config, attribute, "föobar")

        # extra-path has its own negative test

        # invalid target
        with self.assertRaisesMessage(LintConfigError,
                                      "Option target must be an existing directory (current value: 'föo/bar')"):
            config.target = "föo/bar"

    def test_ignore_independent_from_rules(self):
        # Test that the lintconfig rules are not modified when setting config.ignore
        # This was different in the past, this test is mostly here to catch regressions
        config = LintConfig()
        original_rules = config.rules
        config.ignore = ["T1", "T2"]
        self.assertEqual(config.ignore, ["T1", "T2"])
        self.assertSequenceEqual(config.rules, original_rules)

    def test_config_equality(self):
        self.assertEqual(LintConfig(), LintConfig())
        self.assertNotEqual(LintConfig(), LintConfigGenerator())

        # Ensure LintConfig are not equal if they differ on their attributes
        attrs = [("verbosity", 1), ("rules", []), ("ignore_stdin", True), ("debug", True),
                 ("ignore", ["T1"]), ("staged", True), ("_config_path", self.get_sample_path()),
                 ("ignore_merge_commits", False), ("ignore_fixup_commits", False),
                 ("ignore_squash_commits", False), ("ignore_revert_commits", False),
                 ("extra_path", self.get_sample_path("user_rules")), ("target", self.get_sample_path()),
                 ("contrib", ["CC1"])]
        for attr, val in attrs:
            config = LintConfig()
            setattr(config, attr, val)
            self.assertNotEqual(LintConfig(), config)

        # Other attributes don't matter
        config1 = LintConfig()
        config2 = LintConfig()
        config1.foo = "bår"
        self.assertEqual(config1, config2)
        config2.foo = "dūr"
        self.assertEqual(config1, config2)


class LintConfigGeneratorTests(BaseTestCase):
    @staticmethod
    @patch('gitlint.config.shutil.copyfile')
    def test_install_commit_msg_hook_negative(copy):
        LintConfigGenerator.generate_config("föo/bar/test")
        copy.assert_called_with(GITLINT_CONFIG_TEMPLATE_SRC_PATH, "föo/bar/test")



# -*- coding: utf-8 -*-

from gitlint.rules import CommitRule, RuleViolation, ConfigurationRule
from gitlint.options import IntOption, StrOption, ListOption


class GitContextRule(CommitRule):
    """ Rule that tests whether we can correctly access certain gitcontext properties """
    name = "gïtcontext"
    id = "UC1"

    def validate(self, commit):
        violations = [
            RuleViolation(self.id, f"GitContext.current_branch: {commit.context.current_branch}", line_nr=1),
            RuleViolation(self.id, f"GitContext.commentchar: {commit.context.commentchar}", line_nr=1)
        ]

        return violations


class GitCommitRule(CommitRule):
    """ Rule that tests whether we can correctly access certain commit properties """
    name = "gïtcommit"
    id = "UC2"

    def validate(self, commit):
        violations = [
            RuleViolation(self.id, f"GitCommit.branches: {commit.branches}", line_nr=1),
            RuleViolation(self.id, f"GitCommit.custom_prop: {commit.custom_prop}", line_nr=1),
        ]

        return violations


class GitlintConfigurationRule(ConfigurationRule):
    """ Rule that tests whether we can correctly access the config as well as modify the commit message """
    name = "cönfigrule"
    id = "UC3"

    def apply(self, config, commit):
        # We add a line to the commit message body that pulls a value from config, this proves we can modify the body
        # and read the config contents
        commit.message.body.append(f"{config.target} ")  # trailing whitespace deliberate to trigger violation

        # We set a custom property that we access in CommitRule, to prove we can add extra properties to the commit
        commit.custom_prop = "foöbar"

        # We also ignore some extra rules, proving that we can modify the config
        config.ignore.append("B4")


class ConfigurableCommitRule(CommitRule):
    """ Rule that tests that we can add configuration to user-defined rules """
    name = "configürable"
    id = "UC4"

    options_spec = [IntOption("int-öption", 2, "int-öption description"),
                    StrOption("str-öption", "föo", "int-öption description"),
                    ListOption("list-öption", ["foo", "bar"], "list-öption description")]

    def validate(self, _):
        violations = [
            RuleViolation(self.id, f"int-öption: {self.options[u'int-öption'].value}", line_nr=1),
            RuleViolation(self.id, f"str-öption: {self.options[u'str-öption'].value}", line_nr=1),
            RuleViolation(self.id, f"list-öption: {self.options[u'list-öption'].value}", line_nr=1),
        ]

        return violations



# -*- coding: utf-8 -*-

import contextlib
import copy
import io
import logging
import os
import re
import shutil
import tempfile

import unittest

from unittest.mock import patch

from gitlint.git import GitContext
from gitlint.utils import LOG_FORMAT, DEFAULT_ENCODING


class BaseTestCase(unittest.TestCase):
    """ Base class of which all gitlint unit test classes are derived. Provides a number of convenience methods. """

    # In case of assert failures, print the full error message
    maxDiff = None

    SAMPLES_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "samples")
    EXPECTED_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "expected")
    GITLINT_USE_SH_LIB = os.environ.get("GITLINT_USE_SH_LIB", "[NOT SET]")

    def setUp(self):
        self.logcapture = LogCapture()
        self.logcapture.setFormatter(logging.Formatter(LOG_FORMAT))
        logging.getLogger('gitlint').setLevel(logging.DEBUG)
        logging.getLogger('gitlint').handlers = [self.logcapture]

        # Make sure we don't propagate anything to child loggers, we need to do this explicitly here
        # because if you run a specific test file like test_lint.py, we won't be calling the setupLogging() method
        # in gitlint.cli that normally takes care of this
        logging.getLogger('gitlint').propagate = False

    @staticmethod
    @contextlib.contextmanager
    def tempdir():
        tmpdir = tempfile.mkdtemp()
        try:
            yield tmpdir
        finally:
            shutil.rmtree(tmpdir)

    @staticmethod
    def get_sample_path(filename=""):
        # Don't join up empty files names because this will add a trailing slash
        if filename == "":
            return BaseTestCase.SAMPLES_DIR

        return os.path.join(BaseTestCase.SAMPLES_DIR, filename)

    @staticmethod
    def get_sample(filename=""):
        """ Read and return the contents of a file in gitlint/tests/samples """
        sample_path = BaseTestCase.get_sample_path(filename)
        with io.open(sample_path, encoding=DEFAULT_ENCODING) as content:
            sample = content.read()
        return sample

    @staticmethod
    def patch_input(side_effect):
        """ Patches the built-in input() with a provided side-effect """
        module_path = "builtins.input"
        patched_module = patch(module_path, side_effect=side_effect)
        return patched_module

    @staticmethod
    def get_expected(filename="", variable_dict=None):
        """ Utility method to read an expected file from gitlint/tests/expected and return it as a string.
        Optionally replace template variables specified by variable_dict. """
        expected_path = os.path.join(BaseTestCase.EXPECTED_DIR, filename)
        with io.open(expected_path, encoding=DEFAULT_ENCODING) as content:
            expected = content.read()

        if variable_dict:
            expected = expected.format(**variable_dict)
        return expected

    @staticmethod
    def get_user_rules_path():
        return os.path.join(BaseTestCase.SAMPLES_DIR, "user_rules")

    @staticmethod
    def gitcontext(commit_msg_str, changed_files=None, ):
        """ Utility method to easily create gitcontext objects based on a given commit msg string and an optional set of
        changed files"""
        with patch("gitlint.git.git_commentchar") as comment_char:
            comment_char.return_value = "#"
            gitcontext = GitContext.from_commit_msg(commit_msg_str)
            commit = gitcontext.commits[-1]
            if changed_files:
                commit.changed_files = changed_files
            return gitcontext

    @staticmethod
    def gitcommit(commit_msg_str, changed_files=None, **kwargs):
        """ Utility method to easily create git commit given a commit msg string and an optional set of changed files"""
        gitcontext = BaseTestCase.gitcontext(commit_msg_str, changed_files)
        commit = gitcontext.commits[-1]
        for attr, value in kwargs.items():
            setattr(commit, attr, value)
        return commit

    def assert_logged(self, expected):
        """ Asserts that the logs match an expected string or list.
            This method knows how to compare a passed list of log lines as well as a newline concatenated string
            of all loglines. """
        if isinstance(expected, list):
            self.assertListEqual(self.logcapture.messages, expected)
        else:
            self.assertEqual("\n".join(self.logcapture.messages), expected)

    def assert_log_contains(self, line):
        """ Asserts that a certain line is in the logs """
        self.assertIn(line, self.logcapture.messages)

    def assertRaisesRegex(self, expected_exception, expected_regex, *args, **kwargs):
        """ Pass-through method to unittest.TestCase.assertRaisesRegex that applies re.escape() to the passed
            `expected_regex`. This is useful to automatically escape all file paths that might be present in the regex.
        """
        return super().assertRaisesRegex(expected_exception, re.escape(expected_regex), *args, **kwargs)

    def clearlog(self):
        """ Clears the log capture """
        self.logcapture.clear()

    @contextlib.contextmanager
    def assertRaisesMessage(self, expected_exception, expected_msg):  # pylint: disable=invalid-name
        """ Asserts an exception has occurred with a given error message """
        try:
            yield
        except expected_exception as exc:
            exception_msg = str(exc)
            if exception_msg != expected_msg:
                error = f"Right exception, wrong message:\n      got: {exception_msg}\n expected: {expected_msg}"
                raise self.fail(error)
            # else: everything is fine, just return
            return
        except Exception as exc:
            raise self.fail(f"Expected '{expected_exception.__name__}' got '{exc.__class__.__name__}'")

        # No exception raised while we expected one
        raise self.fail(f"Expected to raise {expected_exception.__name__}, didn't get an exception at all")

    def object_equality_test(self, obj, attr_list, ctor_kwargs=None):
        """ Helper function to easily implement object equality tests.
            Creates an object clone for every passed attribute and checks for (in)equality
            of the original object with the clone based on those attributes' values.
            This function assumes all attributes in `attr_list` can be passed to the ctor of `obj.__class__`.
        """
        if not ctor_kwargs:
            ctor_kwargs = {}

        attr_kwargs = {}
        for attr in attr_list:
            attr_kwargs[attr] = getattr(obj, attr)

        # For every attr, clone the object and assert the clone and the original object are equal
        # Then, change the current attr and assert objects are unequal
        for attr in attr_list:
            attr_kwargs_copy = copy.deepcopy(attr_kwargs)
            attr_kwargs_copy.update(ctor_kwargs)
            clone = obj.__class__(**attr_kwargs_copy)
            self.assertEqual(obj, clone)

            # Change attribute and assert objects are different (via both attribute set and ctor)
            setattr(clone, attr, "föo")
            self.assertNotEqual(obj, clone)
            attr_kwargs_copy[attr] = "föo"

            self.assertNotEqual(obj, obj.__class__(**attr_kwargs_copy))


class LogCapture(logging.Handler):
    """ Mock logging handler used to capture any log messages during tests."""

    def __init__(self, *args, **kwargs):
        logging.Handler.__init__(self, *args, **kwargs)
        self.messages = []

    def emit(self, record):
        self.messages.append(self.format(record))

    def clear(self):
        self.messages = []



# -*- coding: utf-8 -*-
# pylint: disable=too-many-function-args,unexpected-keyword-arg
import os
from qa.shell import git, gitlint
from qa.base import BaseTestCase


class HookTests(BaseTestCase):
    """ Integration tests for gitlint commitmsg hooks"""

    VIOLATIONS = ['gitlint: checking commit message...\n',
                  '1: T3 Title has trailing punctuation (.): "WIP: This ïs a title."\n',
                  '1: T5 Title contains the word \'WIP\' (case-insensitive): "WIP: This ïs a title."\n',
                  '2: B4 Second line is not empty: "Contënt on the second line"\n',
                  '3: B6 Body message is missing\n',
                  '-----------------------------------------------\n',
                  'gitlint: \x1b[31mYour commit message contains violations.\x1b[0m\n']

    def setUp(self):
        super().setUp()
        self.responses = []
        self.response_index = 0
        self.githook_output = []

        # The '--staged' flag used in the commit-msg hook fetches additional information from the underlying
        # git repo which means there already needs to be a commit in the repo
        # (as gitlint --staged doesn't work against empty repos)
        self.create_simple_commit("Commït Title\n\nCommit Body explaining commit.")

        # install git commit-msg hook and assert output
        output_installed = gitlint("install-hook", _cwd=self.tmp_git_repo)
        expected_installed = ("Successfully installed gitlint commit-msg hook in "
                              f"{self.tmp_git_repo}/.git/hooks/commit-msg\n")

        self.assertEqualStdout(output_installed, expected_installed)

    def tearDown(self):
        # uninstall git commit-msg hook and assert output
        output_uninstalled = gitlint("uninstall-hook", _cwd=self.tmp_git_repo)
        expected_uninstalled = ("Successfully uninstalled gitlint commit-msg hook from "
                                f"{self.tmp_git_repo}/.git/hooks/commit-msg\n")

        self.assertEqualStdout(output_uninstalled, expected_uninstalled)
        super().tearDown()

    def _violations(self):
        # Make a copy of the violations array so that we don't inadvertently edit it in the test (like I did :D)
        return list(self.VIOLATIONS)

        # callback function that captures git commit-msg hook output

    def _interact(self, line, stdin):
        self.githook_output.append(line)
        # Answer 'yes' to question to keep violating commit-msg
        if "Your commit message contains violations" in line:
            response = self.responses[self.response_index]
            stdin.put(f"{response}\n")
            self.response_index = (self.response_index + 1) % len(self.responses)

    def test_commit_hook_no_violations(self):
        test_filename = self.create_simple_commit("This ïs a title\n\nBody contënt that should work",
                                                  out=self._interact, tty_in=True)

        short_hash = self.get_last_commit_short_hash()
        expected_output = ["gitlint: checking commit message...\n",
                           "gitlint: \x1b[32mOK\x1b[0m (no violations in commit message)\n",
                           f"[master {short_hash}] This ïs a title\n",
                           " 1 file changed, 0 insertions(+), 0 deletions(-)\n",
                           f" create mode 100644 {test_filename}\n"]
        self.assertListEqual(expected_output, self.githook_output)

    def test_commit_hook_continue(self):
        self.responses = ["y"]
        test_filename = self.create_simple_commit("WIP: This ïs a title.\nContënt on the second line",
                                                  out=self._interact, tty_in=True)

        # Determine short commit-msg hash, needed to determine expected output
        short_hash = self.get_last_commit_short_hash()

        expected_output = self._violations()
        expected_output += ["Continue with commit anyways (this keeps the current commit message)? " +
                            "[y(es)/n(no)/e(dit)] " +
                            f"[master {short_hash}] WIP: This ïs a title. Contënt on the second line\n",
                            " 1 file changed, 0 insertions(+), 0 deletions(-)\n",
                            f" create mode 100644 {test_filename}\n"]

        assert len(self.githook_output) == len(expected_output)
        for output, expected in zip(self.githook_output, expected_output):
            self.assertMultiLineEqual(
                output.replace('\r', ''),
                expected.replace('\r', ''))

    def test_commit_hook_abort(self):
        self.responses = ["n"]
        test_filename = self.create_simple_commit("WIP: This ïs a title.\nContënt on the second line",
                                                  out=self._interact, ok_code=1, tty_in=True)
        git("rm", "-f", test_filename, _cwd=self.tmp_git_repo)

        # Determine short commit-msg hash, needed to determine expected output

        expected_output = self._violations()
        expected_output += ["Continue with commit anyways (this keeps the current commit message)? " +
                            "[y(es)/n(no)/e(dit)] " +
                            "Commit aborted.\n",
                            "Your commit message: \n",
                            "-----------------------------------------------\n",
                            "WIP: This ïs a title.\n",
                            "Contënt on the second line\n",
                            "-----------------------------------------------\n"]

        self.assertListEqual(expected_output, self.githook_output)

    def test_commit_hook_edit(self):
        self.responses = ["e", "y"]
        env = {"EDITOR": ":"}
        test_filename = self.create_simple_commit("WIP: This ïs a title.\nContënt on the second line",
                                                  out=self._interact, env=env, tty_in=True)
        git("rm", "-f", test_filename, _cwd=self.tmp_git_repo)

        short_hash = git("rev-parse", "--short", "HEAD", _cwd=self.tmp_git_repo, _tty_in=True).replace("\n", "")

        # Determine short commit-msg hash, needed to determine expected output

        expected_output = self._violations()
        expected_output += ['Continue with commit anyways (this keeps the current commit message)? ' +
                            '[y(es)/n(no)/e(dit)] ' + self._violations()[0]]
        expected_output += self._violations()[1:]
        expected_output += ['Continue with commit anyways (this keeps the current commit message)? ' +
                            "[y(es)/n(no)/e(dit)] " +
                            f"[master {short_hash}] WIP: This ïs a title. Contënt on the second line\n",
                            " 1 file changed, 0 insertions(+), 0 deletions(-)\n",
                            f" create mode 100644 {test_filename}\n"]

        assert len(self.githook_output) == len(expected_output)
        for output, expected in zip(self.githook_output, expected_output):
            self.assertMultiLineEqual(
                output.replace('\r', ''),
                expected.replace('\r', ''))

    def test_commit_hook_worktree(self):
        """ Tests that hook installation and un-installation also work in git worktrees.
        Test steps:
            ```sh
            git init <tmpdir>
            cd <tmpdir>
            git worktree add <worktree-tempdir>
            cd <worktree-tempdir>
            gitlint install-hook
            gitlint uninstall-hook
            ```
        """
        tmp_git_repo = self.create_tmp_git_repo()
        self.create_simple_commit("Simple title\n\nContënt in the body", git_repo=tmp_git_repo)

        worktree_dir = self.generate_temp_path()
        self.tmp_git_repos.append(worktree_dir)  # make sure we clean up the worktree afterwards

        git("worktree", "add", worktree_dir, _cwd=tmp_git_repo, _tty_in=True)

        output_installed = gitlint("install-hook", _cwd=worktree_dir)
        expected_hook_path = os.path.join(tmp_git_repo, ".git", "hooks", "commit-msg")
        expected_msg = f"Successfully installed gitlint commit-msg hook in {expected_hook_path}\r\n"
        self.assertEqual(output_installed, expected_msg)

        output_uninstalled = gitlint("uninstall-hook", _cwd=worktree_dir)
        expected_hook_path = os.path.join(tmp_git_repo, ".git", "hooks", "commit-msg")
        expected_msg = f"Successfully uninstalled gitlint commit-msg hook from {expected_hook_path}\r\n"
        self.assertEqual(output_uninstalled, expected_msg)



import re

from gitlint.options import ListOption
from gitlint.rules import CommitMessageTitle, LineRule, RuleViolation

RULE_REGEX = re.compile(r"([^(]+?)(\([^)]+?\))?!?: .+")


class ConventionalCommit(LineRule):
    """ This rule enforces the spec at https://www.conventionalcommits.org/. """

    name = "contrib-title-conventional-commits"
    id = "CT1"
    target = CommitMessageTitle

    options_spec = [
        ListOption(
            "types",
            ["fix", "feat", "chore", "docs", "style", "refactor", "perf", "test", "revert", "ci", "build"],
            "Comma separated list of allowed commit types.",
        )
    ]

    def validate(self, line, _commit):
        violations = []
        match = RULE_REGEX.match(line)

        if not match:
            msg = "Title does not follow ConventionalCommits.org format 'type(optional-scope): description'"
            violations.append(RuleViolation(self.id, msg, line))
        else:
            line_commit_type = match.group(1)
            if line_commit_type not in self.options["types"].value:
                opt_str = ', '.join(self.options['types'].value)
                violations.append(RuleViolation(self.id, f"Title does not start with one of {opt_str}", line))

        return violations




# -*- coding: utf-8 -*-
from gitlint.tests.base import BaseTestCase
from gitlint.rules import RuleViolation
from gitlint.contrib.rules.signedoff_by import SignedOffBy

from gitlint.config import LintConfig


class ContribSignedOffByTests(BaseTestCase):

    def test_enable(self):
        # Test that rule can be enabled in config
        for rule_ref in ['CC1', 'contrib-body-requires-signed-off-by']:
            config = LintConfig()
            config.contrib = [rule_ref]
            self.assertIn(SignedOffBy(), config.rules)

    def test_signedoff_by(self):
        # No violations when 'Signed-off-by' line is present
        rule = SignedOffBy()
        violations = rule.validate(self.gitcommit("Föobar\n\nMy Body\nSigned-off-by: John Smith"))
        self.assertListEqual([], violations)

        # Assert violation when no 'Signed-off-by' line is present
        violations = rule.validate(self.gitcommit("Föobar\n\nMy Body"))
        expected_violation = RuleViolation("CC1", "Body does not contain a 'Signed-off-by' line", line_nr=1)
        self.assertListEqual(violations, [expected_violation])

        # Assert violation when no 'Signed-off-by' in title but not in body
        violations = rule.validate(self.gitcommit("Signed-off-by\n\nFöobar"))
        self.assertListEqual(violations, [expected_violation])



# -*- coding: utf-8 -*-
# pylint: disable=too-many-function-args,unexpected-keyword-arg
import io
import subprocess
from qa.shell import echo, gitlint
from qa.base import BaseTestCase
from qa.utils import DEFAULT_ENCODING


class StdInTests(BaseTestCase):
    """ Integration tests for various STDIN scenarios for gitlint """

    def test_stdin_pipe(self):
        """ Test piping input into gitlint.
            This is the equivalent of doing:
            $ echo "foo" | gitlint
        """
        # NOTE: There is no use in testing this with _tty_in=True, because if you pipe something into a command
        # there never is a TTY connected to stdin (per definition). We're setting _tty_in=False here to be explicit
        # but note that this is always true when piping something into a command.
        output = gitlint(echo("WIP: Pïpe test."),
                         _cwd=self.tmp_git_repo, _tty_in=False, _err_to_out=True, _ok_code=[3])
        self.assertEqualStdout(output, self.get_expected("test_stdin/test_stdin_pipe_1"))

    def test_stdin_pipe_empty(self):
        """ Test the scenario where no TTY is attached and nothing is piped into gitlint. This occurs in
            CI runners like Jenkins and Gitlab, see https://github.com/jorisroovers/gitlint/issues/42 for details.
            This is the equivalent of doing:
            $ echo -n "" | gitlint
        """
        commit_msg = "WIP: This ïs a title.\nContent on the sëcond line"
        self.create_simple_commit(commit_msg)

        # We need to set _err_to_out explicitly for sh to merge stdout and stderr output in case there's
        # no TTY attached to STDIN
        # http://amoffat.github.io/sh/sections/special_arguments.html?highlight=_tty_in#err-to-out
        output = gitlint(echo("-n", ""), _cwd=self.tmp_git_repo, _tty_in=False, _err_to_out=True, _ok_code=[3])

        self.assertEqual(output, self.get_expected("test_stdin/test_stdin_pipe_empty_1"))

    def test_stdin_file(self):
        """ Test the scenario where STDIN is a regular file (stat.S_ISREG = True)
            This is the equivalent of doing:
            $ gitlint < myfile
        """
        tmp_commit_msg_file = self.create_tmpfile("WIP: STDIN ïs a file test.")

        with io.open(tmp_commit_msg_file, encoding=DEFAULT_ENCODING) as file_handle:

            # We need to use subprocess.Popen() here instead of sh because when passing a file_handle to sh, it will
            # deal with reading the file itself instead of passing it on to gitlint as a STDIN. Since we're trying to
            # test for the condition where stat.S_ISREG == True that won't work for us here.
            with subprocess.Popen("gitlint", stdin=file_handle, cwd=self.tmp_git_repo,
                                  stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as p:
                output, _ = p.communicate()
                self.assertEqual(output.decode(DEFAULT_ENCODING), self.get_expected("test_stdin/test_stdin_file_1"))



# -*- coding: utf-8 -*-

from gitlint.rules import ConfigurationRule
from gitlint.options import IntOption


"""
Full details on user-defined rules: https://jorisroovers.com/gitlint/user_defined_rules

The ReleaseConfigurationRule class below is an example of a user-defined ConfigurationRule. Configuration rules are
gitlint rules that are applied once per commit and BEFORE any other rules are run. Configuration Rules are meant to
dynamically change gitlint's configuration and/or the commit that is about to be linted. A typically use-case for this
is modifying the behavior of gitlint's rules based on a commit contents.

Notes:
- Modifying the commit object DOES NOT modify the actual git commit message in the target repo, only gitlint's copy of
  it.
- Modifying the config object only has effect on the commit that is being linted, subsequent commits will not
  automatically inherit this configuration.
"""


class ReleaseConfigurationRule(ConfigurationRule):
    """
    This rule will modify gitlint's behavior for Release Commits.

    This example might not be the most realistic for a real-world scenario,
    but is meant to give an overview of what's possible.
    """

    # A rule MUST have a human friendly name
    name = "release-configuration-rule"

    # A rule MUST have a *unique* id, we recommend starting with UCR
    # (for User-defined Configuration-Rule), but this can really be anything.
    id = "UCR1"

    # A rule MAY have an option_spec if its behavior should be configurable.
    options_spec = [IntOption('custom-verbosity', 2, "Gitlint verbosity for release commits")]

    def apply(self, config, commit):
        self.log.debug("ReleaseConfigurationRule: This will be visible when running `gitlint --debug`")

        # If the commit title starts with 'Release', we want to modify
        # how all subsequent rules interpret that commit
        if commit.message.title.startswith("Release"):

            # If your Release commit messages are auto-generated, the
            # body might contain trailing whitespace. Let's ignore that
            config.ignore.append("body-trailing-whitespace")

            # Similarly, the body lines might exceed 80 chars,
            # let's set gitlint's limit to 200
            # To set rule options use:
            # config.set_rule_option(<rule-name>, <rule-option>, <value>)
            config.set_rule_option("body-max-line-length", "line-length", 200)

            # For kicks, let's set gitlint's verbosity to 2
            # To set general options use
            # config.set_general_option(<general-option>, <value>)
            config.set_general_option("verbosity", 2)
            # Wwe can also use custom options to make this configurable
            config.set_general_option("verbosity", self.options['custom-verbosity'].value)

            # Strip any lines starting with $ from the commit message
            # (this only affects how gitlint sees your commit message, it does
            # NOT modify your actual commit in git)
            commit.message.body = [line for line in commit.message.body if not line.startswith("$")]

            # You can add any extra properties you want to the commit object, these will be available later on
            # in all rules.
            commit.my_property = "This is my property"



# -*- coding: utf-8 -*-

import io
from io import StringIO
import os

from click.testing import CliRunner

from unittest.mock import patch

from gitlint.tests.base import BaseTestCase
from gitlint import cli
from gitlint import hooks
from gitlint import config
from gitlint.shell import ErrorReturnCode

from gitlint.utils import DEFAULT_ENCODING


class CLIHookTests(BaseTestCase):
    USAGE_ERROR_CODE = 253
    GIT_CONTEXT_ERROR_CODE = 254
    CONFIG_ERROR_CODE = 255

    def setUp(self):
        super(CLIHookTests, self).setUp()
        self.cli = CliRunner()

        # Patch gitlint.cli.git_version() so that we don't have to patch it separately in every test
        self.git_version_path = patch('gitlint.cli.git_version')
        cli.git_version = self.git_version_path.start()
        cli.git_version.return_value = "git version 1.2.3"

    def tearDown(self):
        self.git_version_path.stop()

    @patch('gitlint.hooks.GitHookInstaller.install_commit_msg_hook')
    @patch('gitlint.hooks.git_hooks_dir', return_value=os.path.join("/hür", "dur"))
    def test_install_hook(self, _, install_hook):
        """ Test for install-hook subcommand """
        result = self.cli.invoke(cli.cli, ["install-hook"])
        expected_path = os.path.join("/hür", "dur", hooks.COMMIT_MSG_HOOK_DST_PATH)
        expected = f"Successfully installed gitlint commit-msg hook in {expected_path}\n"
        self.assertEqual(result.output, expected)
        self.assertEqual(result.exit_code, 0)
        expected_config = config.LintConfig()
        expected_config.target = os.path.realpath(os.getcwd())
        install_hook.assert_called_once_with(expected_config)

    @patch('gitlint.hooks.GitHookInstaller.install_commit_msg_hook')
    @patch('gitlint.hooks.git_hooks_dir', return_value=os.path.join("/hür", "dur"))
    def test_install_hook_target(self, _, install_hook):
        """  Test for install-hook subcommand with a specific --target option specified """
        # Specified target
        result = self.cli.invoke(cli.cli, ["--target", self.SAMPLES_DIR, "install-hook"])
        expected_path = os.path.join("/hür", "dur", hooks.COMMIT_MSG_HOOK_DST_PATH)
        expected = "Successfully installed gitlint commit-msg hook in %s\n" % expected_path
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, expected)

        expected_config = config.LintConfig()
        expected_config.target = self.SAMPLES_DIR
        install_hook.assert_called_once_with(expected_config)

    @patch('gitlint.hooks.GitHookInstaller.install_commit_msg_hook', side_effect=hooks.GitHookInstallerError("tëst"))
    def test_install_hook_negative(self, install_hook):
        """ Negative test for install-hook subcommand """
        result = self.cli.invoke(cli.cli, ["install-hook"])
        self.assertEqual(result.exit_code, self.GIT_CONTEXT_ERROR_CODE)
        self.assertEqual(result.output, "tëst\n")
        expected_config = config.LintConfig()
        expected_config.target = os.path.realpath(os.getcwd())
        install_hook.assert_called_once_with(expected_config)

    @patch('gitlint.hooks.GitHookInstaller.uninstall_commit_msg_hook')
    @patch('gitlint.hooks.git_hooks_dir', return_value=os.path.join("/hür", "dur"))
    def test_uninstall_hook(self, _, uninstall_hook):
        """ Test for uninstall-hook subcommand """
        result = self.cli.invoke(cli.cli, ["uninstall-hook"])
        expected_path = os.path.join("/hür", "dur", hooks.COMMIT_MSG_HOOK_DST_PATH)
        expected = f"Successfully uninstalled gitlint commit-msg hook from {expected_path}\n"
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, expected)
        expected_config = config.LintConfig()
        expected_config.target = os.path.realpath(os.getcwd())
        uninstall_hook.assert_called_once_with(expected_config)

    @patch('gitlint.hooks.GitHookInstaller.uninstall_commit_msg_hook', side_effect=hooks.GitHookInstallerError("tëst"))
    def test_uninstall_hook_negative(self, uninstall_hook):
        """ Negative test for uninstall-hook subcommand """
        result = self.cli.invoke(cli.cli, ["uninstall-hook"])
        self.assertEqual(result.exit_code, self.GIT_CONTEXT_ERROR_CODE)
        self.assertEqual(result.output, "tëst\n")
        expected_config = config.LintConfig()
        expected_config.target = os.path.realpath(os.getcwd())
        uninstall_hook.assert_called_once_with(expected_config)

    def test_run_hook_no_tty(self):
        """ Test for run-hook subcommand.
            When no TTY is available (like is the case for this test), the hook will abort after the first check.
        """

        # No need to patch git as we're passing a msg-filename to run-hook, so no git calls are made.
        # Note that this is the case when passing --staged as well, but that's tested as part of the integration tests
        # (=end-to-end scenario).

        # Ideally we'd be able to assert that run-hook internally calls the lint cli command, but couldn't make
        # that work. Have tried many different variatons of mocking and patching without avail. For now, we just
        # check the output which indirectly proves the same thing.

        with self.tempdir() as tmpdir:
            msg_filename = os.path.join(tmpdir, "hür")
            with io.open(msg_filename, 'w', encoding=DEFAULT_ENCODING) as f:
                f.write("WIP: tïtle\n")

            with patch('gitlint.display.stderr', new=StringIO()) as stderr:
                result = self.cli.invoke(cli.cli, ["--msg-filename", msg_filename, "run-hook"])
                self.assertEqual(result.output, self.get_expected('cli/test_cli_hooks/test_hook_no_tty_1_stdout'))
                self.assertEqual(stderr.getvalue(), self.get_expected("cli/test_cli_hooks/test_hook_no_tty_1_stderr"))

                # exit code is 1 because aborted (no stdin available)
                self.assertEqual(result.exit_code, 1)

    @patch('gitlint.cli.shell')
    def test_run_hook_edit(self, shell):
        """ Test for run-hook subcommand, answering 'e(dit)' after commit-hook """

        set_editors = [None, "myeditor"]
        expected_editors = ["vim -n", "myeditor"]
        commit_messages = ["WIP: höok edit 1", "WIP: höok edit 2"]

        for i in range(0, len(set_editors)):
            if set_editors[i]:
                os.environ['EDITOR'] = set_editors[i]

            with self.patch_input(['e', 'e', 'n']):
                with self.tempdir() as tmpdir:
                    msg_filename = os.path.realpath(os.path.join(tmpdir, "hür"))
                    with io.open(msg_filename, 'w', encoding=DEFAULT_ENCODING) as f:
                        f.write(commit_messages[i] + "\n")

                    with patch('gitlint.display.stderr', new=StringIO()) as stderr:
                        result = self.cli.invoke(cli.cli, ["--msg-filename", msg_filename, "run-hook"])
                        self.assertEqual(result.output, self.get_expected('cli/test_cli_hooks/test_hook_edit_1_stdout',
                                                                          {"commit_msg": commit_messages[i]}))
                        expected = self.get_expected("cli/test_cli_hooks/test_hook_edit_1_stderr",
                                                     {"commit_msg": commit_messages[i]})
                        self.assertEqual(stderr.getvalue(), expected)

                        # exit code = number of violations
                        self.assertEqual(result.exit_code, 2)

                        shell.assert_called_with(expected_editors[i] + " " + msg_filename)
                        self.assert_log_contains("DEBUG: gitlint.cli run-hook: editing commit message")
                        self.assert_log_contains(f"DEBUG: gitlint.cli run-hook: {expected_editors[i]} {msg_filename}")

    def test_run_hook_no(self):
        """ Test for run-hook subcommand, answering 'n(o)' after commit-hook """

        with self.patch_input(['n']):
            with self.tempdir() as tmpdir:
                msg_filename = os.path.join(tmpdir, "hür")
                with io.open(msg_filename, 'w', encoding=DEFAULT_ENCODING) as f:
                    f.write("WIP: höok no\n")

                with patch('gitlint.display.stderr', new=StringIO()) as stderr:
                    result = self.cli.invoke(cli.cli, ["--msg-filename", msg_filename, "run-hook"])
                    self.assertEqual(result.output, self.get_expected('cli/test_cli_hooks/test_hook_no_1_stdout'))
                    self.assertEqual(stderr.getvalue(), self.get_expected("cli/test_cli_hooks/test_hook_no_1_stderr"))

                    # We decided not to keep the commit message: hook returns number of violations (>0)
                    # This will cause git to abort the commit
                    self.assertEqual(result.exit_code, 2)
                    self.assert_log_contains("DEBUG: gitlint.cli run-hook: commit message declined")

    def test_run_hook_yes(self):
        """ Test for run-hook subcommand, answering 'y(es)' after commit-hook """
        with self.patch_input(['y']):
            with self.tempdir() as tmpdir:
                msg_filename = os.path.join(tmpdir, "hür")
                with io.open(msg_filename, 'w', encoding=DEFAULT_ENCODING) as f:
                    f.write("WIP: höok yes\n")

                with patch('gitlint.display.stderr', new=StringIO()) as stderr:
                    result = self.cli.invoke(cli.cli, ["--msg-filename", msg_filename, "run-hook"])
                    self.assertEqual(result.output, self.get_expected('cli/test_cli_hooks/test_hook_yes_1_stdout'))
                    self.assertEqual(stderr.getvalue(), self.get_expected("cli/test_cli_hooks/test_hook_yes_1_stderr"))

                    # Exit code is 0 because we decide to keep the commit message
                    # This will cause git to keep the commit
                    self.assertEqual(result.exit_code, 0)
                    self.assert_log_contains("DEBUG: gitlint.cli run-hook: commit message accepted")

    @patch('gitlint.cli.get_stdin_data', return_value=False)
    @patch('gitlint.git.sh')
    def test_run_hook_negative(self, sh, _):
        """ Negative test for the run-hook subcommand: testing whether exceptions are correctly handled when
        running `gitlint run-hook`.
        """
        # GIT_CONTEXT_ERROR_CODE: git error
        error_msg = b"fatal: not a git repository (or any of the parent directories): .git"
        sh.git.side_effect = ErrorReturnCode("full command", b"stdout", error_msg)
        result = self.cli.invoke(cli.cli, ["run-hook"])
        expected = self.get_expected('cli/test_cli_hooks/test_run_hook_negative_1', {'git_repo': os.getcwd()})
        self.assertEqual(result.output, expected)
        self.assertEqual(result.exit_code, self.GIT_CONTEXT_ERROR_CODE)

        # USAGE_ERROR_CODE: incorrect use of gitlint
        result = self.cli.invoke(cli.cli, ["--staged", "run-hook"])
        self.assertEqual(result.output, self.get_expected('cli/test_cli_hooks/test_run_hook_negative_2'))
        self.assertEqual(result.exit_code, self.USAGE_ERROR_CODE)

        # CONFIG_ERROR_CODE: incorrect config. Note that this is handled before the hook even runs
        result = self.cli.invoke(cli.cli, ["-c", "föo.bár=1", "run-hook"])
        self.assertEqual(result.output, "Config Error: No such rule 'föo'\n")
        self.assertEqual(result.exit_code, self.CONFIG_ERROR_CODE)

    @patch('gitlint.cli.get_stdin_data', return_value="WIP: Test hook stdin tïtle\n")
    def test_run_hook_stdin_violations(self, _):
        """ Test for passing stdin data to run-hook, expecting some violations. Equivalent of:
            $ echo "WIP: Test hook stdin tïtle" | gitlint run-hook
        """

        with patch('gitlint.display.stderr', new=StringIO()) as stderr:
            result = self.cli.invoke(cli.cli, ["run-hook"])
            expected_stderr = self.get_expected('cli/test_cli_hooks/test_hook_stdin_violations_1_stderr')
            self.assertEqual(stderr.getvalue(), expected_stderr)
            self.assertEqual(result.output, self.get_expected('cli/test_cli_hooks/test_hook_stdin_violations_1_stdout'))
            # Hook will auto-abort because we're using stdin. Abort = exit code 1
            self.assertEqual(result.exit_code, 1)

    @patch('gitlint.cli.get_stdin_data', return_value="Test tïtle\n\nTest bödy that is long enough")
    def test_run_hook_stdin_no_violations(self, _):
        """ Test for passing stdin data to run-hook, expecting *NO* violations, Equivalent of:
            $ echo -e "Test tïtle\n\nTest bödy that is long enough" | gitlint run-hook
        """

        with patch('gitlint.display.stderr', new=StringIO()) as stderr:
            result = self.cli.invoke(cli.cli, ["run-hook"])
            self.assertEqual(stderr.getvalue(), "")  # no errors = no stderr output
            expected_stdout = self.get_expected('cli/test_cli_hooks/test_hook_stdin_no_violations_1_stdout')
            self.assertEqual(result.output, expected_stdout)
            self.assertEqual(result.exit_code, 0)

    @patch('gitlint.cli.get_stdin_data', return_value="WIP: Test hook config tïtle\n")
    def test_run_hook_config(self, _):
        """ Test that gitlint still respects config when running run-hook, equivalent of:
            $ echo "WIP: Test hook config tïtle" | gitlint -c title-max-length.line-length=5 --ignore B6 run-hook
        """

        with patch('gitlint.display.stderr', new=StringIO()) as stderr:
            result = self.cli.invoke(cli.cli, ["-c", "title-max-length.line-length=5", "--ignore", "B6", "run-hook"])
            self.assertEqual(stderr.getvalue(), self.get_expected('cli/test_cli_hooks/test_hook_config_1_stderr'))
            self.assertEqual(result.output, self.get_expected('cli/test_cli_hooks/test_hook_config_1_stdout'))
            # Hook will auto-abort because we're using stdin. Abort = exit code 1
            self.assertEqual(result.exit_code, 1)

    @patch('gitlint.cli.get_stdin_data', return_value=False)
    @patch('gitlint.git.sh')
    def test_run_hook_local_commit(self, sh, _):
        """ Test running the hook on the last commit-msg from the local repo, equivalent of:
            $ gitlint run-hook
            and then choosing 'e'
        """
        sh.git.side_effect = [
            "6f29bf81a8322a04071bb794666e48c443a90360",
            "test åuthor\x00test-email@föo.com\x002016-12-03 15:28:15 +0100\x00åbc\n"
            "WIP: commït-title\n\ncommït-body",
            "#",  # git config --get core.commentchar
            "commit-1-branch-1\ncommit-1-branch-2\n",
            "file1.txt\npåth/to/file2.txt\n"
        ]

        with self.patch_input(['e']):
            with patch('gitlint.display.stderr', new=StringIO()) as stderr:
                result = self.cli.invoke(cli.cli, ["run-hook"])
                expected = self.get_expected('cli/test_cli_hooks/test_hook_local_commit_1_stderr')
                self.assertEqual(stderr.getvalue(), expected)
                self.assertEqual(result.output, self.get_expected('cli/test_cli_hooks/test_hook_local_commit_1_stdout'))
                # If we can't edit the message, run-hook follows regular gitlint behavior and exit code = # violations
                self.assertEqual(result.exit_code, 2)



# -*- coding: utf-8 -*-
import copy

from gitlint.tests.base import BaseTestCase

from gitlint.config import LintConfig, LintConfigBuilder, LintConfigError

from gitlint import rules


class LintConfigBuilderTests(BaseTestCase):
    def test_set_option(self):
        config_builder = LintConfigBuilder()
        config = config_builder.build()

        # assert some defaults
        self.assertEqual(config.get_rule_option('title-max-length', 'line-length'), 72)
        self.assertEqual(config.get_rule_option('body-max-line-length', 'line-length'), 80)
        self.assertListEqual(config.get_rule_option('title-must-not-contain-word', 'words'), ["WIP"])
        self.assertEqual(config.verbosity, 3)

        # Make some changes and check blueprint
        config_builder.set_option('title-max-length', 'line-length', 100)
        config_builder.set_option('general', 'verbosity', 2)
        config_builder.set_option('title-must-not-contain-word', 'words', ["foo", "bar"])
        expected_blueprint = {'title-must-not-contain-word': {'words': ['foo', 'bar']},
                              'title-max-length': {'line-length': 100}, 'general': {'verbosity': 2}}
        self.assertDictEqual(config_builder._config_blueprint, expected_blueprint)

        # Build config and verify that the changes have occurred and no other changes
        config = config_builder.build()
        self.assertEqual(config.get_rule_option('title-max-length', 'line-length'), 100)
        self.assertEqual(config.get_rule_option('body-max-line-length', 'line-length'), 80)  # should be unchanged
        self.assertListEqual(config.get_rule_option('title-must-not-contain-word', 'words'), ["foo", "bar"])
        self.assertEqual(config.verbosity, 2)

    def test_set_from_commit_ignore_all(self):
        config = LintConfig()
        original_rules = config.rules
        original_rule_ids = [rule.id for rule in original_rules]

        config_builder = LintConfigBuilder()

        # nothing gitlint
        config_builder.set_config_from_commit(self.gitcommit("tëst\ngitlint\nfoo"))
        config = config_builder.build()
        self.assertSequenceEqual(config.rules, original_rules)
        self.assertListEqual(config.ignore, [])

        # ignore all rules
        config_builder.set_config_from_commit(self.gitcommit("tëst\ngitlint-ignore: all\nfoo"))
        config = config_builder.build()
        self.assertEqual(config.ignore, original_rule_ids)

        # ignore all rules, no space
        config_builder.set_config_from_commit(self.gitcommit("tëst\ngitlint-ignore:all\nfoo"))
        config = config_builder.build()
        self.assertEqual(config.ignore, original_rule_ids)

        # ignore all rules, more spacing
        config_builder.set_config_from_commit(self.gitcommit("tëst\ngitlint-ignore: \t all\nfoo"))
        config = config_builder.build()
        self.assertEqual(config.ignore, original_rule_ids)

    def test_set_from_commit_ignore_specific(self):
        # ignore specific rules
        config_builder = LintConfigBuilder()
        config_builder.set_config_from_commit(self.gitcommit("tëst\ngitlint-ignore: T1, body-hard-tab"))
        config = config_builder.build()
        self.assertEqual(config.ignore, ["T1", "body-hard-tab"])

    def test_set_from_config_file(self):
        # regular config file load, no problems
        config_builder = LintConfigBuilder()
        config_builder.set_from_config_file(self.get_sample_path("config/gitlintconfig"))
        config = config_builder.build()

        # Do some assertions on the config
        self.assertEqual(config.verbosity, 1)
        self.assertFalse(config.debug)
        self.assertFalse(config.ignore_merge_commits)
        self.assertIsNone(config.extra_path)
        self.assertEqual(config.ignore, ["title-trailing-whitespace", "B2"])

        self.assertEqual(config.get_rule_option('title-max-length', 'line-length'), 20)
        self.assertEqual(config.get_rule_option('body-max-line-length', 'line-length'), 30)

    def test_set_from_config_file_negative(self):
        config_builder = LintConfigBuilder()

        # bad config file load
        foo_path = self.get_sample_path("föo")
        expected_error_msg = f"Invalid file path: {foo_path}"
        with self.assertRaisesMessage(LintConfigError, expected_error_msg):
            config_builder.set_from_config_file(foo_path)

        # error during file parsing
        path = self.get_sample_path("config/no-sections")
        expected_error_msg = "File contains no section headers."
        # We only match the start of the message here, since the exact message can vary depending on platform
        with self.assertRaisesRegex(LintConfigError, expected_error_msg):
            config_builder.set_from_config_file(path)

        # non-existing rule
        path = self.get_sample_path("config/nonexisting-rule")
        config_builder = LintConfigBuilder()
        config_builder.set_from_config_file(path)
        expected_error_msg = "No such rule 'föobar'"
        with self.assertRaisesMessage(LintConfigError, expected_error_msg):
            config_builder.build()

        # non-existing general option
        path = self.get_sample_path("config/nonexisting-general-option")
        config_builder = LintConfigBuilder()
        config_builder.set_from_config_file(path)
        expected_error_msg = "'foo' is not a valid gitlint option"
        with self.assertRaisesMessage(LintConfigError, expected_error_msg):
            config_builder.build()

        # non-existing option
        path = self.get_sample_path("config/nonexisting-option")
        config_builder = LintConfigBuilder()
        config_builder.set_from_config_file(path)
        expected_error_msg = "Rule 'title-max-length' has no option 'föobar'"
        with self.assertRaisesMessage(LintConfigError, expected_error_msg):
            config_builder.build()

        # invalid option value
        path = self.get_sample_path("config/invalid-option-value")
        config_builder = LintConfigBuilder()
        config_builder.set_from_config_file(path)
        expected_error_msg = "'föo' is not a valid value for option 'title-max-length.line-length'. " + \
                             "Option 'line-length' must be a positive integer (current value: 'föo')."
        with self.assertRaisesMessage(LintConfigError, expected_error_msg):
            config_builder.build()

    def test_set_config_from_string_list(self):
        config = LintConfig()

        # change and assert changes
        config_builder = LintConfigBuilder()
        config_builder.set_config_from_string_list(['general.verbosity=1', 'title-max-length.line-length=60',
                                                    'body-max-line-length.line-length=120',
                                                    "title-must-not-contain-word.words=håha"])

        config = config_builder.build()
        self.assertEqual(config.get_rule_option('title-max-length', 'line-length'), 60)
        self.assertEqual(config.get_rule_option('body-max-line-length', 'line-length'), 120)
        self.assertListEqual(config.get_rule_option('title-must-not-contain-word', 'words'), ["håha"])
        self.assertEqual(config.verbosity, 1)

    def test_set_config_from_string_list_negative(self):
        config_builder = LintConfigBuilder()

        # assert error on incorrect rule - this happens at build time
        config_builder.set_config_from_string_list(["föo.bar=1"])
        with self.assertRaisesMessage(LintConfigError, "No such rule 'föo'"):
            config_builder.build()

        # no equal sign
        expected_msg = "'föo.bar' is an invalid configuration option. Use '<rule>.<option>=<value>'"
        with self.assertRaisesMessage(LintConfigError, expected_msg):
            config_builder.set_config_from_string_list(["föo.bar"])

        # missing value
        expected_msg = "'föo.bar=' is an invalid configuration option. Use '<rule>.<option>=<value>'"
        with self.assertRaisesMessage(LintConfigError, expected_msg):
            config_builder.set_config_from_string_list(["föo.bar="])

        # space instead of equal sign
        expected_msg = "'föo.bar 1' is an invalid configuration option. Use '<rule>.<option>=<value>'"
        with self.assertRaisesMessage(LintConfigError, expected_msg):
            config_builder.set_config_from_string_list(["föo.bar 1"])

        # no period between rule and option names
        expected_msg = "'föobar=1' is an invalid configuration option. Use '<rule>.<option>=<value>'"
        with self.assertRaisesMessage(LintConfigError, expected_msg):
            config_builder.set_config_from_string_list([u'föobar=1'])

    def test_rebuild_config(self):
        # normal config build
        config_builder = LintConfigBuilder()
        config_builder.set_option('general', 'verbosity', 3)
        lint_config = config_builder.build()
        self.assertEqual(lint_config.verbosity, 3)

        # check that existing config gets overwritten when we pass it to a configbuilder with different options
        existing_lintconfig = LintConfig()
        existing_lintconfig.verbosity = 2
        lint_config = config_builder.build(existing_lintconfig)
        self.assertEqual(lint_config.verbosity, 3)
        self.assertEqual(existing_lintconfig.verbosity, 3)

    def test_clone(self):
        config_builder = LintConfigBuilder()
        config_builder.set_option('general', 'verbosity', 2)
        config_builder.set_option('title-max-length', 'line-length', 100)
        expected = {'title-max-length': {'line-length': 100}, 'general': {'verbosity': 2}}
        self.assertDictEqual(config_builder._config_blueprint, expected)

        # Clone and verify that the blueprint is the same as the original
        cloned_builder = config_builder.clone()
        self.assertDictEqual(cloned_builder._config_blueprint, expected)

        # Modify the original and make sure we're not modifying the clone (i.e. check that the copy is a deep copy)
        config_builder.set_option('title-max-length', 'line-length', 120)
        self.assertDictEqual(cloned_builder._config_blueprint, expected)

    def test_named_rules(self):
        # Store a copy of the default rules from the config, so we can reference it later
        config_builder = LintConfigBuilder()
        config = config_builder.build()
        default_rules = copy.deepcopy(config.rules)
        self.assertEqual(default_rules, config.rules)  # deepcopy should be equal

        # Add a named rule by setting an option in the config builder that follows the named rule pattern
        # Assert that whitespace in the rule name is stripped
        rule_qualifiers = [u'T7:my-extra-rüle', u' T7 :   my-extra-rüle  ', u'\tT7:\tmy-extra-rüle\t',
                           u'T7:\t\n  \tmy-extra-rüle\t\n\n', "title-match-regex:my-extra-rüle"]
        for rule_qualifier in rule_qualifiers:
            config_builder = LintConfigBuilder()
            config_builder.set_option(rule_qualifier, 'regex', "föo")

            expected_rules = copy.deepcopy(default_rules)
            my_rule = rules.TitleRegexMatches({'regex': "föo"})
            my_rule.id = rules.TitleRegexMatches.id + ":my-extra-rüle"
            my_rule.name = rules.TitleRegexMatches.name + ":my-extra-rüle"
            expected_rules._rules[u'T7:my-extra-rüle'] = my_rule
            self.assertEqual(config_builder.build().rules, expected_rules)

            # assert that changing an option on the newly added rule is passed correctly to the RuleCollection
            # we try this with all different rule qualifiers to ensure they all are normalized and map
            # to the same rule
            for other_rule_qualifier in rule_qualifiers:
                cb = config_builder.clone()
                cb.set_option(other_rule_qualifier, 'regex', other_rule_qualifier + "bōr")
                # before setting the expected rule option value correctly, the RuleCollection should be different
                self.assertNotEqual(cb.build().rules, expected_rules)
                # after setting the option on the expected rule, it should be equal
                my_rule.options['regex'].set(other_rule_qualifier + "bōr")
                self.assertEqual(cb.build().rules, expected_rules)
                my_rule.options['regex'].set("wrong")

    def test_named_rules_negative(self):
        # T7 = title-match-regex
        # Invalid rule name
        for invalid_name in ["", " ", "    ", "\t", "\n", "å b", "å:b", "åb:", ":åb"]:
            config_builder = LintConfigBuilder()
            config_builder.set_option(f"T7:{invalid_name}", 'regex', "tëst")
            expected_msg = f"The rule-name part in 'T7:{invalid_name}' cannot contain whitespace, colons or be empty"
            with self.assertRaisesMessage(LintConfigError, expected_msg):
                config_builder.build()

        # Invalid parent rule name
        config_builder = LintConfigBuilder()
        config_builder.set_option("Ž123:foöbar", "fåke-option", "fåke-value")
        with self.assertRaisesMessage(LintConfigError, "No such rule 'Ž123' (named rule: 'Ž123:foöbar')"):
            config_builder.build()

        # Invalid option name (this is the same as with regular rules)
        config_builder = LintConfigBuilder()
        config_builder.set_option("T7:foöbar", "blå", "my-rëgex")
        with self.assertRaisesMessage(LintConfigError, "Rule 'T7:foöbar' has no option 'blå'"):
            config_builder.build()
